from flask_apispec import use_kwargs as api_spec_use_kwargs
from flask_marshmallow.fields import _rapply, _url_val
from marshmallow import post_load
from marshmallow import fields as ma_fields
from opsy.flask_extensions import ma


def use_kwargs(schema_cls, schema_kwargs=None, **kwargs):

    if schema_kwargs is None:
        schema_kwargs = {}

    def factory(request):
        # Respect partial updates for PATCH and GET requests
        partial = getattr(request, 'method', None) in ['PATCH', 'GET']
        # partial = request.method in ['PATCH', 'GET']
        # Add current request to the schema's context
        return schema_cls(
            partial=partial,
            context={'request': request},
            **schema_kwargs
        )

    return api_spec_use_kwargs(factory, **kwargs)


class Hyperlinks(ma_fields.Dict):
    # We recreate this from upstream, but inherit from Dict so apispec gets
    # the right type.

    _CHECK_ATTRIBUTE = False

    def __init__(self, schema, **kwargs):
        self.schema = schema
        ma_fields.Dict.__init__(self, **kwargs)

    def _serialize(self, value, attr, obj):
        return _rapply(self.schema, _url_val, key=attr, obj=obj)


class Password(ma.String):
    """Field to obscure passwords on serialization."""

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return None
        return '<HIDDEN>'


class EmptySchema(ma.Schema):
    pass


###############################################################################
# Base schemas
###############################################################################


class BaseSchema(ma.ModelSchema):

    @post_load
    def make_instance(self, data, **kwargs):
        """Return deserialized data as a dict, not a model instance."""
        return data



from marshmallow import RAISE
from opsy.flask_extensions import ma
from opsy.schema import BaseSchema, Password
from opsy.monitoring.models import Dashboard, Event, MonitoringService


###############################################################################
# Non-sqlalchemy schemas
###############################################################################

class HttpPollerBackendSchema(BaseSchema):
    class Meta:
        fields = ('host', 'protocol', 'port', 'path', 'interval', 'timeout',
                  'username', 'password', 'verify_ssl')
        ordered = True
        unknown = RAISE

    host = ma.String()
    protocol = ma.String()
    port = ma.Integer()
    path = ma.String()
    interval = ma.Integer()
    timeout = ma.Integer()
    username = ma.String()
    password = Password()
    verify_ssl = ma.Boolean()

###############################################################################
# Sqlalchemy schemas
###############################################################################


class MonitoringServiceSchema(BaseSchema):

    class Meta:
        model = MonitoringService
        fields = ('id', 'name', 'zone_id', 'enabled',
                  'backend_config', 'backend_name', 'interval',
                  'last_poll_time', 'extra')
        ordered = True
        unknown = RAISE

    backend_config = ma.Nested('HttpPollerBackendSchema')


class EventSchema(BaseSchema):

    class Meta:
        model = Event
        fields = ('id', 'monitoring_service_id', 'monitoring_service_name',
                  'zone_name', 'assignee_name', 'host_id'
                  'host_name', 'check_name', 'state', 'status', 'interval',
                  'interval', 'occurrences', 'command', 'output', 'resolved',
                  'resolved_at', 'extra')
        ordered = True
        unknown = RAISE


class DashboardSchema(BaseSchema):

    class Meta:
        model = Dashboard
        fields = ('id', 'client_name', 'check_name', 'zone_name', 'zone_id',
                  'backend', 'client_subscriptions', 'check_subscribers',
                  'occurrences_threshold', 'occurrences', 'status', 'interval',
                  'command', 'output', 'last_poll_time', 'updated_at',
                  'silences')
        ordered = True
        unknown = RAISE



"""Add ldap_user column to user model

Revision ID: b0ee04f071c6
Revises: 22d14acc37df
Create Date: 2019-11-27 16:03:16.477422

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b0ee04f071c6'
down_revision = '22d14acc37df'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('users', sa.Column('ldap_user', sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'ldap_user')
    # ### end Alembic commands ###



from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from flask_allows import Allows
from flask_apispec.extension import FlaskApiSpec
from flask_jsglue import JSGlue
from flask_ldap3_login import LDAP3LoginManager
from flask_login import LoginManager, current_user
from flask_marshmallow import Marshmallow
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from prometheus_flask_exporter import PrometheusMetrics

allows = Allows()  # pylint: disable=invalid-name
db = SQLAlchemy()  # pylint: disable=invalid-name
migrate = Migrate()  # pylint: disable=invalid-name
jsglue = JSGlue()  # pylint: disable=invalid-name
ldap_manager = LDAP3LoginManager()  # pylint: disable=invalid-name
login_manager = LoginManager()  # pylint: disable=invalid-name
ma = Marshmallow()  # pylint: disable=invalid-name
apispec = FlaskApiSpec()  # pylint: disable=invalid-name
metrics = PrometheusMetrics(app=None)  # pylint: disable=invalid-name


# pylint: disable=unused-import
def configure_extensions(app):
    # Make SQLAlchemy aware of models
    from opsy.auth import models as am  # noqa: F401
    from opsy.inventory import models as im  # noqa: F401
    from opsy import __version__ as opsy_version  # noqa: F401
    db.init_app(app)
    migrate.init_app(app, db=db)
    ma.init_app(app)
    jsglue.init_app(app)
    login_manager.init_app(app)
    if app.config.opsy['auth']['ldap_enabled']:
        ldap_manager.init_app(app)
    allows.init_app(app)
    allows.identity_loader(lambda: current_user)
    app.config.update({
        'APISPEC_SPEC': APISpec(
            title='opsy',
            version='v1',
            openapi_version='2.0',
            info={'description': "It's Opsy!"},
            plugins=[MarshmallowPlugin()],
            basePath=app.config.opsy['app']['uri_prefix']
        ),
        'APISPEC_SWAGGER_URL': '/docs/swagger.json',
        'APISPEC_SWAGGER_UI_URL': '/docs/',
    })
    app.config['APISPEC_SPEC'].components.security_scheme(
        'api_key', {'type': 'apiKey', 'in': 'header', 'name': 'X-AUTH-TOKEN'})
    apispec.init_app(app)
    from opsy.auth.utils import (load_user, load_user_from_request,
                                 APISessionInterface)
    login_manager.user_loader(load_user)
    login_manager.request_loader(load_user_from_request)
    app.session_interface = APISessionInterface()
    metrics.init_app(app)
    metrics.info('app_info', 'Application info', version=opsy_version)


def finalize_extensions(app):
    # Workaround for https://github.com/jmcarp/flask-apispec/issues/111
    # pylint: disable=protected-access
    for key, value in apispec.spec._paths.items():
        apispec.spec._paths[key] = {
            inner_key: inner_value
            for inner_key, inner_value in value.items()
            if inner_key != 'options'
        }



from marshmallow import fields as ma_fields
from marshmallow import RAISE
from marshmallow_sqlalchemy import field_for
from opsy.auth.models import User, Role, Permission
from opsy.flask_extensions import ma
from opsy.schema import BaseSchema, Hyperlinks

###############################################################################
# Non-sqlalchemy schemas
###############################################################################


class UserLoginSchema(BaseSchema):

    user_name = ma_fields.String(load_only=True, required=True)
    password = ma_fields.String(load_only=True, required=True)
    force_renew = ma_fields.Boolean(load_only=True, default=False,
                                    missing=False)


class AppPermissionSchema(BaseSchema):
    class Meta:
        fields = ('endpoint', 'method', 'permission_needed')
        ordered = True
        unknown = RAISE

    endpoint = ma_fields.String()
    method = ma_fields.String()
    permission_needed = ma_fields.String()

###############################################################################
# Sqlalchemy schemas
###############################################################################


class UserSchema(BaseSchema):

    class Meta:
        model = User
        fields = ('id', 'name', 'full_name', 'email', 'enabled', 'ldap_user',
                  'created_at', 'updated_at', 'roles', 'permissions', '_links')
        ordered = True
        unknown = RAISE

    id = field_for(User, 'id', dump_only=True)
    name = field_for(User, 'name', required=True)
    email = field_for(
        User, 'email', field_class=ma.Email)  # pylint: disable=no-member
    created_at = field_for(User, 'created_at', dump_only=True)
    updated_at = field_for(User, 'updated_at', dump_only=True)

    permissions = ma.Nested(  # pylint: disable=no-member
        'RolePermissionRefSchema', many=True, dump_only=True)
    roles = ma.Nested(  # pylint: disable=no-member
        'RoleRefSchema', many=True, dump_only=True)
    _links = Hyperlinks(
        {"self": ma.URLFor("auth_users.users_get", id_or_name="<id>"),
         "collection": ma.URLFor("auth_users.users_list")},
        dump_only=True
    )


class UserCreateSchema(UserSchema):
    """
    UserCreateSchema.

    This exists because the password field with load_only=True was still
    being rendered on the output UserSchema. Seems like a bug, but I didn't
    want to fight it.
    """

    class Meta:
        model = User
        fields = ('name', 'full_name', 'email', 'enabled', 'ldap_user',
                  'password')
        ordered = True
        unknown = RAISE

    password = ma_fields.String(required=True, load_only=True)


class UserUpdateSchema(UserSchema):

    class Meta:
        model = User
        fields = ('name', 'full_name', 'email', 'enabled', 'ldap_user',
                  'password')
        ordered = True
        unknown = RAISE

    name = field_for(User, 'name', required=False)
    password = ma_fields.String(load_only=True)


class UserQuerySchema(UserSchema):

    class Meta:
        model = User
        fields = ('id', 'name', 'email', 'permission_name',
                  'role_id', 'role_name', 'enabled', 'ldap_user')
        ordered = True
        unknown = RAISE

    id = field_for(User, 'id')
    name = field_for(User, 'name', required=False)
    email = field_for(User, 'email')
    permission_name = ma_fields.String(attribute='permissions___name')
    role_id = ma_fields.String(attribute='role___id')
    role_name = ma_fields.String(attribute='role___name')
    enabled = field_for(User, 'enabled')


class UserRefSchema(UserSchema):

    class Meta:
        model = User
        fields = ('id', 'name', '_links')
        ordered = True
        unknown = RAISE


class UserTokenSchema(UserSchema):

    class Meta:
        model = User
        fields = ('id', 'name', 'full_name', 'email', 'ldap_user',
                  'session_token', 'session_token_expires_at')
        ordered = True
        unknown = RAISE

    id = field_for(User, 'id', data_key='user_id', dump_only=True)
    name = field_for(User, 'name', data_key='user_name', required=True)
    password = ma_fields.String(required=True, load_only=True)
    session_token = field_for(
        User, 'session_token', data_key='token', dump_only=True)
    session_token_expires_at = field_for(
        User, 'session_token_expires_at', data_key='expires_at',
        dump_only=True)


class UserTokenUpdateSchema(UserSchema):

    class Meta:
        model = User
        fields = ('full_name', 'email', 'password')
        ordered = True
        unknown = RAISE


class RoleSchema(BaseSchema):

    class Meta:
        model = Role
        fields = ('id', 'name', 'ldap_group', 'description', 'created_at',
                  'updated_at', 'permissions', 'users', '_links')
        ordered = True
        unknown = RAISE

    id = field_for(Role, 'id', dump_only=True)
    name = field_for(Role, 'name', required=True)
    created_at = field_for(Role, 'created_at', dump_only=True)
    updated_at = field_for(Role, 'updated_at', dump_only=True)

    permissions = ma.Nested(  # pylint: disable=no-member
        'RolePermissionRefSchema', many=True, dump_only=True)
    users = ma.Nested(  # pylint: disable=no-member
        'UserRefSchema', many=True, dump_only=True)

    _links = Hyperlinks(
        {"self": ma.URLFor("auth_roles.roles_get", id_or_name="<id>"),
         "collection": ma.URLFor("auth_roles.roles_list")},
        dump_only=True)


class RoleUpdateSchema(RoleSchema):

    class Meta:
        model = Role
        fields = ('name', 'ldap_group', 'description')
        ordered = True
        unknown = RAISE

    name = field_for(Role, 'name', required=False)


class RoleQuerySchema(RoleSchema):

    class Meta:
        model = Role
        fields = ('id', 'name', 'ldap_group', 'permssion_name',
                  'user_name', 'user_id')
        ordered = True
        unknown = RAISE

    id = field_for(Role, 'id')
    name = field_for(Role, 'name', required=False)
    ldap_group = field_for(Role, 'ldap_group')
    permission_name = ma_fields.String(attribute='permissions___name')
    user_name = ma_fields.String(attribute='users___name')
    user_id = ma_fields.String(attribute='users___id')


class RoleRefSchema(RoleSchema):

    class Meta:
        model = Role
        fields = ('id', 'name', '_links')
        ordered = True
        unknown = RAISE


class RolePermissionSchema(BaseSchema):

    class Meta:
        model = Permission
        fields = ('id', 'role', 'name', 'created_at', 'updated_at', '_links')
        ordered = True

    id = field_for(Permission, 'id', dump_only=True)
    name = field_for(Permission, 'name', required=True)
    created_at = field_for(Permission, 'created_at', dump_only=True)
    updated_at = field_for(Permission, 'updated_at', dump_only=True)

    role = ma.Nested(  # pylint: disable=no-member
        'RoleRefSchema', dump_only=True)

    _links = Hyperlinks({
        "self": ma.URLFor("auth_roles.role_permissions_get",
                          id_or_name="<role_id>",
                          permission_id_or_name="<id>"),
        "collection": ma.URLFor("auth_roles.role_permissions_list",
                                id_or_name="<role_id>")}, dump_only=True)


class RolePermissionUpdateSchema(BaseSchema):

    class Meta:
        model = Permission
        fields = ('name', )
        ordered = True

    name = field_for(Permission, 'name', required=False)


class RolePermissionQuerySchema(RolePermissionSchema):

    class Meta:
        model = Permission
        fields = ('id', 'name')
        ordered = True
        uknown = RAISE

    id = field_for(Permission, 'id')
    name = field_for(Permission, 'name', required=False)


class RolePermissionRefSchema(RolePermissionSchema):

    class Meta:
        model = Permission
        fields = ('id', 'name', '_links')
        ordered = True
        unknown = RAISE



import pytest
import datetime as dt
from opsy.auth.models import User
from opsy.inventory.models import Host, Group

###############################################################################
# OpsyQuery
###############################################################################


def test_opsy_query(test_user, test_users, test_inventory_bootstrap):
    """Test opsy_query to make sure it works correctly."""
    # Test get_or_fail
    assert User.query.get_or_fail(test_user.id) == test_user
    with pytest.raises(ValueError):
        User.query.get_or_fail('thisisnotarealuser')
    # Test first_or_fail
    assert User.query.filter_by(id=test_user.id).first_or_fail() == test_user
    with pytest.raises(ValueError):
        User.query.filter_by(id='thisisnotarealuser').first_or_fail()
    # Setup for filter_in tests
    westprom = Host.query.filter_by(name='westprom').first()
    westconsul = Host.query.filter_by(name='westconsul').first()
    centralprom = Host.query.filter_by(name='centralprom').first()
    centralconsul = Host.query.filter_by(name='centralconsul').first()
    eastprom = Host.query.filter_by(name='eastprom').first()
    eastconsul = Host.query.filter_by(name='eastconsul').first()
    # Test empty query
    assert len(Host.query.filter_in().all()) == 6
    # Test if we don't give it a string it bypasses the string operations
    eastconsul.update(updated_at=dt.datetime.utcnow())
    assert Host.query.filter_in(
        updated_at=eastconsul.updated_at).first() == eastconsul
    # Test filter_in include
    assert Host.query.filter_in(name='westprom').first() == westprom
    for node in Host.query.filter_in(name='westprom,eastprom').all():
        assert node in [westprom, eastprom]
    # Test filter_in exclude
    assert len(Host.query.filter_in(name='!westprom').all()) == 5
    for node in Host.query.filter_in(name='!westprom').all():
        assert isinstance(node, Host)
        assert node != westprom
        assert len(Host.query.filter_in(name='!westprom,!eastprom').all()) == 4
    for node in Host.query.filter_in(name='!westprom,!eastprom').all():
        assert isinstance(node, Host)
        assert node not in [westprom, eastprom]
    # Test filter_in like
    assert len(Host.query.filter_in(name='*prom').all()) == 3
    for node in Host.query.filter_in(name='*prom').all():
        assert isinstance(node, Host)
        assert node in [westprom, centralprom, eastprom]
    # Test filter_in not like
    assert len(Host.query.filter_in(name='!*prom').all()) == 3
    for node in Host.query.filter_in(name='!*prom').all():
        assert isinstance(node, Host)
        assert node not in [westprom, centralprom, eastprom]
    # Now all of these over again but with relationship filters!
    # We do these with both groups___name and zone___name since they're
    # different types of relationships (many-to-many and one-to-many).
    # Test filter_in include
    assert len(Host.query.filter_in(zone___name='east,west').all()) == 4
    for node in Host.query.filter_in(zone___name='east,west').all():
        assert isinstance(node, Host)
        assert node in [westprom, westconsul, eastprom, eastconsul]
    assert len(Host.query.filter_in(groups___name='prom_nodes').all()) == 3
    for node in Host.query.filter_in(groups___name='prom_nodes').all():
        assert isinstance(node, Host)
        assert node in [westprom, centralprom, eastprom]
    # Test filter_in exclude
    assert len(Host.query.filter_in(zone___name='!west,!east').all()) == 2
    for node in Host.query.filter_in(zone___name='!west,!east').all():
        assert isinstance(node, Host)
        assert node in [centralprom, centralconsul]
    assert len(Host.query.filter_in(groups___name='!prom_nodes').all()) == 3
    for node in Host.query.filter_in(groups___name='!prom_nodes').all():
        assert isinstance(node, Host)
        assert node in [westconsul, centralconsul, eastconsul]
    # Test filter_in like
    assert len(Host.query.filter_in(zone___name='*st').all()) == 4
    for node in Host.query.filter_in(zone___name='*st').all():
        assert isinstance(node, Host)
        assert node in [westprom, westconsul, eastprom, eastconsul]
    assert len(Host.query.filter_in(groups___name='prom*').all()) == 3
    for node in Host.query.filter_in(groups___name='prom*').all():
        assert isinstance(node, Host)
        assert node in [westprom, centralprom, eastprom]
    # Test filter_in not like
    assert len(Host.query.filter_in(zone___name='!*st').all()) == 2
    for node in Host.query.filter_in(zone___name='!*st').all():
        assert isinstance(node, Host)
        assert node in [centralprom, centralconsul]
    assert len(Host.query.filter_in(groups___name='!prom*').all()) == 3
    for node in Host.query.filter_in(groups___name='!prom*').all():
        assert isinstance(node, Host)
        assert node not in [westprom, centralprom, eastprom]


###############################################################################
# BaseModel
###############################################################################


def test_base_model(test_group, test_groups):
    """Test BaseModel to make sure it works correctly."""
    # Test __repr__ (have to use super here since Group overrides this)
    assert super(Group, test_group).__repr__() == \
        f'<{test_group.__class__.__name__} {test_group.id}>'
    # Test create
    new_group = Group.create(name='new_group')
    assert new_group.id is not None
    assert Group.query.filter_by(id=new_group.id).first() == new_group
    # Test update
    new_group.update(name='my_new_group')
    assert new_group.name == 'my_new_group'
    # Test Delete
    new_group_id = new_group.id
    new_group.delete()
    assert Group.query.filter_by(id=new_group_id).first() is None
    # Test get_by_id
    assert Group.get_by_id(test_group.id) == test_group
    with pytest.raises(ValueError):
        Group.get_by_id('thisgroupdoesntexist')
    # Test updated_by_id
    Group.update_by_id(test_group.id, name='my_test_group')
    assert test_group.name == 'my_test_group'
    with pytest.raises(ValueError):
        Group.update_by_id('thisgroupdoesntexist', name='Cool bois')
    # Test delete_by_id
    test_group_id = test_group.id
    Group.delete_by_id(test_group.id)
    assert Group.query.filter_by(id=test_group_id).first() is None
    with pytest.raises(ValueError):
        Group.delete_by_id('thisgroupdoesntexist')


###############################################################################
# NamedModel
###############################################################################


def test_named_model(test_user, test_users):
    """Test NamedModel to make sure it works correctly."""
    # Test __repr__
    assert test_user.__repr__() == \
        f'<{test_user.__class__.__name__} {test_user.name}>'
    # Test create
    new_user = User.create('new_user')
    assert new_user.id is not None
    assert User.query.filter_by(id=new_user.id).first() == new_user
    # Test get_by_id_or_name
    assert User.get_by_id_or_name(new_user.id) == new_user
    assert User.get_by_id_or_name(new_user.name) == new_user
    with pytest.raises(ValueError):
        User.get_by_id_or_name('thisuserdoesntexist')
    # Test update_by_id_or_name
    User.update_by_id_or_name(new_user.name, full_name='Bobby Tables')
    assert new_user.full_name == 'Bobby Tables'
    User.update_by_id_or_name(new_user.id, full_name='Little Bobby Tables')
    assert new_user.full_name == 'Little Bobby Tables'
    with pytest.raises(ValueError):
        User.update_by_id_or_name('thisuserdoesntexist', full_name='Hat Man')
    # Test delete_by_id_or_name
    new_user_id = new_user.id
    test_user_id = test_user.id
    User.delete_by_id_or_name(new_user.id)
    assert User.query.filter_by(id=new_user_id).first() is None
    User.delete_by_id_or_name(test_user.name)
    assert User.query.filter_by(id=test_user_id).first() is None
    with pytest.raises(ValueError):
        User.delete_by_id_or_name('thisuserdoesntexist')



from faker import Faker
from opsy.inventory.models import Zone, Group, Host


def test_zone():
    try:
        return Zone.get_by_id_or_name('test')
    except ValueError:
        return Zone.create(name='test')


def test_zones(number=10):
    fake = Faker(locale='en_US')
    fake.seed(8675309)
    zones = []
    for x in range(0, number):
        zones.append(Zone.create(name=fake.domain_word()))
    return zones


def test_group():
    return Group.create(name='test')


def test_groups(number=10):
    fake = Faker(locale='en_US')
    fake.seed(8675309)
    groups = []
    for x in range(0, number):
        groups.append(Group.create(name=fake.domain_word()))
    return groups


def test_host():
    try:
        zone = Zone.get_by_id_or_name('test')
    except ValueError:
        zone = Zone.create(name='test')
    return Host.create(name='test', zone_id=zone.id)


def test_hosts(number=10):
    fake = Faker(locale='en_US')
    fake.seed(8675309)
    try:
        zone = Zone.get_by_id_or_name('test')
    except ValueError:
        zone = Zone.create(name='test')
    hosts = []
    for x in range(0, number):
        hosts.append(Host.create(name=fake.domain_word(), zone_id=zone.id))
    return hosts


def test_inventory_bootstrap():
    """Just populates the DB with some inventory assets for testing."""
    west_zone = Zone.create(
        name='west', vars={'datacenter': 'west', 'prom_node': 'westprom'})
    central_zone = Zone.create(
        name='central', vars={'datacenter': 'central', 'prom_node': 'centralprom'})
    east_zone = Zone.create(
        name='east', vars={'datacenter': 'east', 'prom_node': 'eastprom'})
    west_default = Group.create(
        name='default', zone=west_zone,
        vars={'shutup': 'meg', 'default': True})
    global_prom = Group.create(name='prom_nodes', vars={'thanos': True})
    west_prom = Group.create(
        name='prom_nodes', zone=west_zone, parent=global_prom,
        vars={'prom_region': 'west', 'shutup': 'sam'})
    central_prom = Group.create(
        name='prom_nodes', zone=central_zone, parent=global_prom,
        vars={'prom_region': 'central', 'thanos': False})
    east_prom = Group.create(
        name='prom_nodes', zone=east_zone, parent=global_prom,
        vars={'prom_region': 'east'})
    for group in [west_prom, central_prom, east_prom]:
        consul = Host.create(
            name=f'{group.zone.name}consul', zone=group.zone)
        prom = Host.create(
            name=f'{group.zone.name}prom', zone=group.zone,
            vars={'consul_node': consul.name})
        prom.add_group(group)
        if group.zone.name == 'west':
            prom.add_group(west_default)
            consul.add_group(west_default)
    return True



import threading
import time
import pytest
from opsy.app import create_app
from opsy.config import load_config
from opsy.server import create_server
from tests.data import auth, inventory
from opsy.flask_extensions import db as opsy_db


###############################################################################
# Flask Fixtures
###############################################################################


@pytest.fixture(scope='session')
def app():
    test_app = create_app(load_config('tests/test_opsy_config.toml'))
    test_app.config['TESTING'] = True
    return test_app


@pytest.fixture(scope='session')
def _db(app):
    opsy_db.app = app
    opsy_db.drop_all()
    opsy_db.create_all()
    opsy_db.session.commit()
    yield opsy_db
    opsy_db.session.remove()
    opsy_db.drop_all()
    opsy_db.session.bind.dispose()


@pytest.fixture(scope='function')
def test_server(app):
    host = app.config.opsy['server']['host']
    port = app.config.opsy['server']['port']
    server = create_server(app, host, port)
    server.shutdown_timeout = 0  # Speed-up tests teardown
    threading.Thread(target=server.safe_start).start()
    while not server.ready:  # wait until fully initialized and bound
        time.sleep(0.1)
    yield server
    server.stop()


###############################################################################
# Auth Fixtures
###############################################################################


@pytest.fixture(scope='function')
def test_user(db_session):
    """Creates a test user."""
    return auth.test_user()


@pytest.fixture(scope='function')
def disabled_user(db_session):
    """Creates a disabled user."""
    return auth.disabled_user()


@pytest.fixture(scope='function')
def admin_user(db_session):
    """Creates a test admin user."""
    return auth.admin_user()


@pytest.fixture(scope='function')
def test_users(db_session):
    """Creates test users."""
    return auth.test_users()


@pytest.fixture(scope='function')
def test_role(db_session):
    """Creates a test role."""
    return auth.test_role()


@pytest.fixture(scope='function')
def test_roles(db_session):
    """Creates test roles."""
    return auth.test_roles()


###############################################################################
# Inventory Fixtures
###############################################################################


@pytest.fixture(scope='function')
def test_zone(db_session):
    """Creates a test zone."""
    return inventory.test_zone()


@pytest.fixture(scope='function')
def test_zones(db_session):
    """Creates test zones."""
    return inventory.test_zones()


@pytest.fixture(scope='function')
def test_group(db_session):
    """Creates a test group."""
    return inventory.test_group()


@pytest.fixture(scope='function')
def test_groups(db_session):
    """Creates test groups."""
    return inventory.test_groups()


@pytest.fixture(scope='function')
def test_host(db_session):
    """Creates a test host."""
    return inventory.test_host()


@pytest.fixture(scope='function')
def test_hosts(db_session):
    """Creates test hosts."""
    return inventory.test_hosts()


@pytest.fixture(scope='function')
def test_inventory_bootstrap(db_session, mocker):
    """Creates a test inventory environment."""

    # This is to workaround the race condition outlined in this issue:
    # https://github.com/jeancochrane/pytest-flask-sqlalchemy/issues/18
    # Seems to only really bite us with the inventory tests, so just putting
    # here.
    mocker.patch.object(db_session, 'remove', lambda: None)
    return inventory.test_inventory_bootstrap()



import json
import pytest
from time import sleep
from opsy.auth.models import User, Role, Permission
from opsy.auth.schema import UserTokenSchema

###############################################################################
# Login Tests
###############################################################################


def test_login_get(client, test_user):
    """Functional test for login_get."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's make sure it yells at us if we don't give it a token
    response = client.get(
        '/api/v1/login/',
        follow_redirects=True)
    assert response.status_code == 403
    # Now let's make sure it gives us our token if we do have a token
    response = client.get(
        '/api/v1/login/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    # Then we make sure we get the proper response codes.
    assert response.status_code == 200
    assert json.loads(response.data) == UserTokenSchema().dump(test_user)


def test_login_post(client, test_user, disabled_user):
    """Functional test for login_post."""
    test_user_id = test_user.id
    disabled_user_id = disabled_user.id
    # Let's try to give it a non-existant user
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'user_name': 'notrealboy', 'password': 'hopeless'})
    assert response.status_code == 401
    # Now let's try a wrong password
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'user_name': 'test', 'password': 'hopeless'})
    assert response.status_code == 401
    # Now no password
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'user_name': 'test'})
    assert response.status_code == 422
    # Now no user_name
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'password': 'hopeless'})
    assert response.status_code == 422
    # Now neither
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={})
    assert response.status_code == 422
    # Let's try to login with disabled_user and make sure it yells at us
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'user_name': 'disabled', 'password': 'banhammer'})
    assert response.status_code == 401
    # Make sure it still has no session token
    disabled_user = User.get_by_id(disabled_user_id)
    assert disabled_user.session_token is None
    # Okay, now lets actually log in
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'user_name': 'test', 'password': 'weakpass'})
    assert response.status_code == 200
    # And make sure it gives us is what it's supposed to
    test_user = User.get_by_id(test_user_id)
    assert json.loads(response.data) == UserTokenSchema().dump(test_user)
    assert json.loads(response.data)['token'] == test_user.session_token

    # Let's save the token it gave us
    old_token = json.loads(response.data)['token']
    # sleep 1 second just to make sure any timestamps are different for token
    # generation
    sleep(1)
    # login again to make sure it gives us the old, but still valid token
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'user_name': 'test', 'password': 'weakpass'})
    assert json.loads(response.data)['token'] == old_token
    # now let's pass force_renew to make sure it gives us a fresh token
    response = client.post(
        '/api/v1/login/',
        follow_redirects=True,
        json={'force_renew': True,
              'user_name': 'test',
              'password': 'weakpass'})
    assert json.loads(response.data)['token'] != old_token


def test_login_patch(client, test_user):
    """Functional test for login_patch."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's try to patch without a token
    response = client.patch(
        '/api/v1/login/',
        follow_redirects=True,
        json={'full_name': 'I am a test',
              'email': 'user@example.com',
              'password': 'howdypassword'})
    assert response.status_code == 403
    # Now let's make sure we can patch with the token.
    response = client.patch(
        '/api/v1/login/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)],
        json={'full_name': 'I am a test',
              'email': 'user@example.com',
              'password': 'howdypassword'})
    assert response.status_code == 200
    assert json.loads(response.data)['full_name'] == 'I am a test'
    assert json.loads(response.data)['email'] == 'user@example.com'
    assert test_user.verify_password('howdypassword') is True
    # Now let's make sure it yells if we're an LDAP user
    test_user.update(ldap_user=True)
    test_user_token = test_user.get_id()
    response = client.patch(
        '/api/v1/login/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)],
        json={'full_name': 'I am a test',
              'email': 'user@example.com',
              'password': 'howdypassword'})
    assert response.status_code == 400


def test_login_delete(client, test_user):
    """Functional test for login_post."""
    # First lets get our token
    test_user_id = test_user.id
    test_user_token = test_user.get_id()
    # Let's try to logout without a token
    response = client.delete(
        '/api/v1/login/',
        follow_redirects=True)
    assert response.status_code == 403
    # Now let's try with a wrong token
    response = client.delete(
        '/api/v1/login/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', 'notrealtoken')])
    assert response.status_code == 403
    # Okay, now lets actually log out
    response = client.delete(
        '/api/v1/login/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 205
    # And make sure it actually logged us out
    test_user = User.get_by_id(test_user_id)
    assert test_user.session_token is None
    assert test_user.session_token_expires_at is None

###############################################################################
# User Tests
###############################################################################


def test_users_list(client, test_user, test_users):
    """Functional test for users_list."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's make sure it yells at us if we don't give it a token
    response = client.get(
        '/api/v1/users/',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.get(
        '/api/v1/users/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('list_users')
    response = client.get(
        '/api/v1/users/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # We should have 11 users.
    assert len(json.loads(response.data)) == 11
    # Now let's try with a filter
    response = client.get(
        f'/api/v1/users/?name={test_user.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # We should have 1 user.
    assert len(json.loads(response.data)) == 1
    assert json.loads(response.data)[0]['name'] == test_user.name


def test_users_post(client, test_user):
    """Functional test for users_post."""
    user_data = {
        'name': 'example',
        'full_name': 'Example User',
        'email': 'user@example.com',
        'enabled': True,
        'ldap_user': False,
        'password': 'myshinypassword'}
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's make sure it yells at us if we don't give it a token
    response = client.post(
        '/api/v1/users/',
        follow_redirects=True,
        json=user_data)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.post(
        '/api/v1/users/',
        follow_redirects=True,
        json=user_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('create_user')
    response = client.post(
        '/api/v1/users/',
        follow_redirects=True,
        json=user_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 201
    assert json.loads(response.data)['name'] == user_data['name']
    assert json.loads(response.data)['full_name'] == user_data['full_name']
    assert json.loads(response.data)['email'] == user_data['email']
    assert json.loads(response.data)['enabled'] == user_data['enabled']
    assert json.loads(response.data)['ldap_user'] == user_data['ldap_user']
    # Let's make sure it didn't short change us and actually made our user
    example_user = User.get_by_id_or_name('example')
    assert example_user.name == user_data['name']
    assert example_user.full_name == user_data['full_name']
    assert example_user.email == user_data['email']
    assert example_user.enabled == user_data['enabled']
    assert example_user.ldap_user == user_data['ldap_user']
    assert example_user.verify_password(user_data['password']) is True
    # Now try again to make sure it yells at us since the username is taken
    response = client.post(
        '/api/v1/users/',
        follow_redirects=True,
        json=user_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 400


def test_users_get(client, test_user, test_users):
    """Functional test for users_get."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our generated users for these tests
    example_user = test_users[0]
    # Let's make sure it yells at us if we don't give it a token
    response = client.get(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.get(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('show_user')
    response = client.get(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # Make sure we got our user
    assert json.loads(response.data)['id'] == example_user.id
    assert json.loads(response.data)['name'] == example_user.name
    # Make sure it yells at us if the user is bogus
    response = client.get(
        f'/api/v1/users/fake_user',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_users_patch(client, test_user, test_users):
    """Functional test for users_patch."""
    user_data = {
        'name': 'example',
        'full_name': 'Example User',
        'email': 'user@example.com',
        'enabled': True,
        'ldap_user': False,
        'password': 'myshinypassword'}
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our generated users for these tests
    example_user = test_users[0]
    # Let's make sure it yells at us if we don't give it a token
    response = client.patch(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        json=user_data)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.patch(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        json=user_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('update_user')
    response = client.patch(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        json=user_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # Make sure the return data is right
    assert json.loads(response.data)['name'] == user_data['name']
    assert json.loads(response.data)['full_name'] == user_data['full_name']
    assert json.loads(response.data)['email'] == user_data['email']
    assert json.loads(response.data)['enabled'] == user_data['enabled']
    assert json.loads(response.data)['ldap_user'] == user_data['ldap_user']
    # Make sure it actually updated the user
    assert example_user.name == user_data['name']
    assert example_user.full_name == user_data['full_name']
    assert example_user.email == user_data['email']
    assert example_user.enabled == user_data['enabled']
    assert example_user.ldap_user == user_data['ldap_user']
    assert example_user.verify_password(user_data['password']) is True
    # Make sure it yells at us if the user is bogus
    response = client.patch(
        f'/api/v1/users/fake_user',
        follow_redirects=True,
        json=user_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_users_delete(client, test_user, test_users):
    """Functional test for users_delete."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our generated users for these tests
    example_user = test_users[0]
    # Let's go ahead and grab the ID now so we can verify it got nuked later
    example_user_id = example_user.id
    # Let's make sure it yells at us if we don't give it a token
    response = client.delete(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.delete(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('delete_user')
    response = client.delete(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 204
    # Make sure the user is actually gone
    with pytest.raises(ValueError):
        User.get_by_id_or_name(example_user_id)
    # Trying again should give us a 404 since the user doesn't exist
    response = client.delete(
        f'/api/v1/users/{example_user.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


###############################################################################
# Role Tests
###############################################################################


def test_roles_list(client, test_user, test_role, test_roles):
    """Functional test for roles_list."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's make sure it yells at us if we don't give it a token
    response = client.get(
        '/api/v1/roles/',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.get(
        '/api/v1/roles/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('list_roles')
    response = client.get(
        '/api/v1/roles/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # We should have 12 roles.
    assert len(json.loads(response.data)) == 12
    # Now let's try with a filter
    response = client.get(
        f'/api/v1/roles/?name={test_role.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # We should have 1 role.
    assert len(json.loads(response.data)) == 1
    assert json.loads(response.data)[0]['name'] == test_role.name


def test_roles_post(client, test_user):
    """Functional test for roles_post."""
    role_data = {
        'name': 'example',
        'ldap_group': 'example',
        'description': 'The rolest role that ever roled.'}
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's make sure it yells at us if we don't give it a token
    response = client.post(
        '/api/v1/roles/',
        follow_redirects=True,
        json=role_data)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.post(
        '/api/v1/roles/',
        follow_redirects=True,
        json=role_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('create_role')
    response = client.post(
        '/api/v1/roles/',
        follow_redirects=True,
        json=role_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 201
    assert json.loads(response.data)['name'] == role_data['name']
    assert json.loads(response.data)['ldap_group'] == role_data['ldap_group']
    assert json.loads(response.data)['description'] == role_data['description']
    # Let's make sure it didn't short change us and actually made our role
    example_role = Role.get_by_id_or_name('example')
    assert example_role.name == role_data['name']
    assert example_role.ldap_group == role_data['ldap_group']
    assert example_role.description == role_data['description']
    # Now try again to make sure it yells at us since the role name is taken
    response = client.post(
        '/api/v1/roles/',
        follow_redirects=True,
        json=role_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 400


def test_roles_get(client, test_user, test_roles):
    """Functional test for roles_get."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our generated roles for these tests
    example_role = test_roles[0]
    # Let's make sure it yells at us if we don't give it a token
    response = client.get(
        f'/api/v1/roles/{example_role.id}',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.get(
        f'/api/v1/roles/{example_role.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('show_role')
    response = client.get(
        f'/api/v1/roles/{example_role.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # Make sure we got our role
    assert json.loads(response.data)['id'] == example_role.id
    assert json.loads(response.data)['name'] == example_role.name
    # Make sure it yells at us if the role is bogus
    response = client.get(
        f'/api/v1/roles/fake_role',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_roles_patch(client, test_user, test_roles):
    """Functional test for roles_patch."""
    role_data = {
        'name': 'example',
        'ldap_group': 'example',
        'description': 'The rolest role that ever roled.'}
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our generated roles for these tests
    example_role = test_roles[0]
    # Let's make sure it yells at us if we don't give it a token
    response = client.patch(
        f'/api/v1/roles/{example_role.id}',
        follow_redirects=True,
        json=role_data)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.patch(
        f'/api/v1/roles/{example_role.id}',
        follow_redirects=True,
        json=role_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('update_role')
    response = client.patch(
        f'/api/v1/roles/{example_role.id}',
        follow_redirects=True,
        json=role_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # Make sure the return data is right
    assert json.loads(response.data)['name'] == role_data['name']
    assert json.loads(response.data)['ldap_group'] == role_data['ldap_group']
    assert json.loads(response.data)['description'] == role_data['description']
    # Make sure it actually updated the role
    assert example_role.name == role_data['name']
    assert example_role.ldap_group == role_data['ldap_group']
    assert example_role.description == role_data['description']
    # Make sure it yells at us if the role is bogus
    response = client.patch(
        f'/api/v1/roles/fake_user',
        follow_redirects=True,
        json=role_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_roles_delete(client, test_user, test_role):
    """Functional test for roles_delete."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's go ahead and grab the ID now so we can verify it got nuked later
    test_role_id = test_role.id
    # Let's make sure it yells at us if we don't give it a token
    response = client.delete(
        f'/api/v1/roles/{test_role.id}',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.delete(
        f'/api/v1/roles/{test_role.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('delete_role')
    response = client.delete(
        f'/api/v1/roles/{test_role.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 204
    # Make sure the role is actually gone
    with pytest.raises(ValueError):
        Role.get_by_id_or_name(test_role_id)
    # Trying again should give us a 404 since the role doesn't exist
    response = client.delete(
        f'/api/v1/roles/{test_role.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


###############################################################################
# Role Permission Tests
###############################################################################


def test_role_permissions_list(client, test_user, test_role):
    """Functional test for role_permissions_list."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's make sure it yells at us if we don't give it a token
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('show_role')
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # We should have 2 permissions.
    assert len(json.loads(response.data)) == 2
    # Now let's try with a filter
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/?name=test1',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # We should have 1 permission.
    assert len(json.loads(response.data)) == 1
    assert json.loads(response.data)[0]['name'] == 'test1'
    # Check to make sure it yells at us if the role is bogus
    response = client.get(
        f'/api/v1/roles/fake_role/permissions/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_role_permissions_post(client, test_user, test_role):
    """Functional test for role_permissions_post."""
    permission_data = {'name': 'permission-to-kill-9'}
    # First lets get our token
    test_user_token = test_user.get_id()
    # Let's make sure it yells at us if we don't give it a token
    response = client.post(
        f'/api/v1/roles/{test_role.id}/permissions/',
        follow_redirects=True,
        json=permission_data)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.post(
        f'/api/v1/roles/{test_role.id}/permissions/',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('update_role')
    response = client.post(
        f'/api/v1/roles/{test_role.id}/permissions/',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 201
    assert json.loads(response.data)['name'] == permission_data['name']
    # Let's make sure it didn't short change us
    example_permission = Permission.query.filter_by(
        role_id=test_role.id, name=permission_data['name']).first()
    assert example_permission.name == permission_data['name']
    # Now try again to make sure it yells at us since the role name is taken
    response = client.post(
        f'/api/v1/roles/{test_role.id}/permissions/',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 400
    # Check to make sure it yells at us if the role is bogus
    response = client.post(
        f'/api/v1/roles/fake_role/permissions/',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_role_permissions_get(client, test_user, test_role):
    """Functional test for role_permissions_get."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our generated permissions for these tests
    test_permission = test_role.permissions[0]
    # Let's make sure it yells at us if we don't give it a token
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('show_role')
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    assert json.loads(response.data)['id'] == test_permission.id
    assert json.loads(response.data)['name'] == test_permission.name
    # Should also work by id
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    assert json.loads(response.data)['id'] == test_permission.id
    assert json.loads(response.data)['name'] == test_permission.name
    # Make sure it yells at us if the permission is bogus
    response = client.get(
        f'/api/v1/roles/{test_role.id}/permissions/fake_permission',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404
    # Make sure it yells at us if the role is bogus
    response = client.get(
        f'/api/v1/roles/fake_role/permissions/fake_permission',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_role_permissions_patch(client, test_user, test_role):
    """Functional test for role_permissions_patch."""
    permission_data = {'name': 'permission-to-kill-9'}
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our generated roles for these tests
    test_permission = test_role.permissions[0]
    # Let's make sure it yells at us if we don't give it a token
    response = client.patch(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True,
        json=permission_data)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.patch(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('update_role')
    response = client.patch(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 200
    # Make sure the return data is right
    assert json.loads(response.data)['name'] == permission_data['name']
    # Make sure it actually updated the permission
    assert test_permission.name == permission_data['name']
    # Make sure it yells at us if the permission is bogus
    response = client.patch(
        f'/api/v1/roles/{test_role.id}/permissions/fake_permission',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404
    # Check to make sure it also yells at us if the role is bogus
    response = client.patch(
        f'/api/v1/roles/fake_role/permissions/fake_permission',
        follow_redirects=True,
        json=permission_data,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404


def test_role_permissions_delete(client, test_user, test_role):
    """Functional test for role_permissions_delete."""
    # First lets get our token
    test_user_token = test_user.get_id()
    # Grab one of our test permissions
    test_permission = test_role.permissions[0]
    # Let's go ahead and grab the ID now so we can verify it got nuked later
    test_permission_id = test_permission.id
    assert Permission.get_by_id(test_permission_id) == test_permission
    # Let's make sure it yells at us if we don't give it a token
    response = client.delete(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True)
    assert response.status_code == 403
    # And with a user with no permissions
    response = client.delete(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 403
    # Now let's add the right permission and try again.
    users_role = Role.get_by_id_or_name('users')
    users_role.add_permission('update_role')
    response = client.delete(
        f'/api/v1/roles/{test_role.id}/permissions/{test_permission.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 204
    # Make sure the permission is actually gone
    with pytest.raises(ValueError):
        Permission.get_by_id(test_permission_id)
    # Make sure it yells at us if the permission is bogus
    response = client.delete(
        f'/api/v1/roles/{test_role.id}/permissions/fake_permission',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404
    # Check to make sure it also yells at us if the role is bogus
    response = client.delete(
        f'/api/v1/roles/fake_role/permissions/fake_permission',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user_token)])
    assert response.status_code == 404



from flask import abort, Blueprint
from flask_allows import requires
from flask_apispec import marshal_with, doc
from flask_login import current_user
from opsy.auth.schema import (
    UserTokenSchema, UserTokenUpdateSchema, UserLoginSchema,
    UserSchema, UserCreateSchema, UserUpdateSchema, UserQuerySchema,
    RoleSchema, RoleUpdateSchema, RoleQuerySchema,
    RolePermissionSchema, RolePermissionUpdateSchema, RolePermissionQuerySchema)
from opsy.auth.models import User, Role, Permission
from opsy.auth.utils import login, logout, create_token
from opsy.flask_extensions import apispec
from opsy.schema import use_kwargs, EmptySchema
from opsy.rbac import is_logged_in, need_permission
from opsy.exceptions import DuplicateError


def create_auth_views(app):
    app.register_blueprint(login_blueprint, url_prefix='/api/v1/login')
    app.register_blueprint(users_blueprint, url_prefix='/api/v1/users')
    app.register_blueprint(roles_blueprint, url_prefix='/api/v1/roles')
    apispec.spec.tag(
        {'name': 'login',
         'description': 'Endpoints for logging in and out of Opsy.'})
    for view in [login_get, login_post, login_patch, login_delete]:
        apispec.register(view, blueprint='auth_login')
    apispec.spec.tag(
        {'name': 'users',
         'description': 'Endpoints for user management.'})
    for view in [users_list, users_post, users_get, users_patch, users_delete]:
        apispec.register(view, blueprint='auth_users')
    apispec.spec.tag(
        {'name': 'roles',
         'description': 'Endpoints for role management.'})
    for view in [roles_list, roles_post, roles_get, roles_patch, roles_delete,
                 role_permissions_list, role_permissions_post,
                 role_permissions_get, role_permissions_patch,
                 role_permissions_delete]:
        apispec.register(view, blueprint='auth_roles')

###############################################################################
# Blueprints
###############################################################################


# pylint: disable=invalid-name
login_blueprint = Blueprint('auth_login', __name__)
# pylint: disable=invalid-name
users_blueprint = Blueprint('auth_users', __name__)
# pylint: disable=invalid-name
roles_blueprint = Blueprint('auth_roles', __name__)

###############################################################################
# Login Views
###############################################################################


@login_blueprint.route('/', methods=['GET'])
@marshal_with(UserTokenSchema(), code=200)
@doc(
    operationId='show_login',
    summary='Get auth token.',
    description='',
    tags=['login'],
    security=[{'api_key': []}])
@requires(is_logged_in)
def login_get():
    create_token(current_user)
    return current_user


@login_blueprint.route('/', methods=['POST'])
@use_kwargs(UserLoginSchema)
@marshal_with(UserTokenSchema(), code=200)
@doc(
    operationId='create_login',
    summary='Login.',
    description='',
    tags=['login'])
def login_post(**kwargs):
    user = login(**kwargs)
    if not user:
        abort(401, 'Username or password incorrect.')
    return current_user


@login_blueprint.route('/', methods=['PATCH'])
@use_kwargs(UserTokenUpdateSchema)
@marshal_with(UserTokenSchema(), code=200)
@doc(
    operationId='patch_login',
    summary='User self-service, such as updating email or password.',
    description='',
    tags=['login'],
    security=[{'api_key': []}])
@requires(is_logged_in)
def login_patch(**kwargs):
    if current_user.ldap_user:
        abort(400, 'This is not available for LDAP users.')
    return current_user.update(**kwargs)


@login_blueprint.route('/', methods=['DELETE'])
@marshal_with(EmptySchema, code=205)
@doc(
    operationId='delete_login',
    summary='Logout.',
    description='',
    tags=['login'],
    security=[{'api_key': []}])
@requires(is_logged_in)
def login_delete():
    logout(current_user)
    return '', 205

###############################################################################
# User Views
###############################################################################


@users_blueprint.route('/', methods=['GET'])
@use_kwargs(UserQuerySchema, locations=['query'])
@marshal_with(UserSchema(many=True), code=200)
@doc(
    operationId='list_users',
    summary='List users.',
    description='',
    tags=['users'],
    security=[{'api_key': []}])
@need_permission('list_users')
def users_list(**kwargs):
    return User.query.filter_in(**kwargs).all()


@users_blueprint.route('/', methods=['POST'])
@use_kwargs(UserCreateSchema)
@marshal_with(UserSchema(), code=201)
@doc(
    operationId='create_user',
    summary='Create a new user.',
    description='',
    tags=['users'],
    security=[{'api_key': []}])
@need_permission('create_user')
def users_post(**kwargs):
    try:
        return User.create(**kwargs), 201
    except (DuplicateError, ValueError) as error:
        abort(400, str(error))


@users_blueprint.route('/<id_or_name>', methods=['GET'])
@marshal_with(UserSchema(), code=200)
@doc(
    operationId='show_user',
    summary='Show a user.',
    description='',
    tags=['users'],
    security=[{'api_key': []}])
@need_permission('show_user')
def users_get(id_or_name):
    try:
        return User.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))


@users_blueprint.route('/<id_or_name>', methods=['PATCH'])
@use_kwargs(UserUpdateSchema)
@marshal_with(UserSchema(), code=200)
@doc(
    operationId='update_user',
    summary='Update a user.',
    description='',
    tags=['users'],
    security=[{'api_key': []}])
@need_permission('update_user')
def users_patch(id_or_name, **kwargs):
    try:
        return User.update_by_id_or_name(id_or_name, **kwargs)
    except ValueError as error:
        abort(404, str(error))


@users_blueprint.route('/<id_or_name>', methods=['DELETE'])
@marshal_with(EmptySchema, code=204)
@doc(
    operationId='delete_user',
    summary='Delete a user.',
    description='',
    tags=['users'],
    security=[{'api_key': []}])
@need_permission('delete_user')
def users_delete(id_or_name):
    try:
        User.delete_by_id_or_name(id_or_name)
        return ('', 204)
    except ValueError as error:
        abort(404, str(error))

###############################################################################
# Role Views
###############################################################################


@roles_blueprint.route('/', methods=['GET'])
@use_kwargs(RoleQuerySchema, locations=['query'])
@marshal_with(RoleSchema(many=True), code=200)
@doc(
    operationId='list_roles',
    summary='List roles.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('list_roles')
def roles_list(**kwargs):
    return Role.query.filter_in(**kwargs).all()


@roles_blueprint.route('/', methods=['POST'])
@use_kwargs(RoleSchema)
@marshal_with(RoleSchema(), code=201)
@doc(
    operationId='create_role',
    summary='Create a new role.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('create_role')
def roles_post(**kwargs):
    try:
        return Role.create(**kwargs), 201
    except (DuplicateError, ValueError) as error:
        abort(400, str(error))


@roles_blueprint.route('/<id_or_name>', methods=['GET'])
@marshal_with(RoleSchema(), code=200)
@doc(
    operationId='show_role',
    summary='Show a role.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('show_role')
def roles_get(id_or_name):
    try:
        return Role.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))


@roles_blueprint.route('/<id_or_name>', methods=['PATCH'])
@use_kwargs(RoleUpdateSchema)
@marshal_with(RoleSchema(), code=200)
@doc(
    operationId='update_role',
    summary='Update a role.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('update_role')
def roles_patch(id_or_name, **kwargs):
    try:
        return Role.update_by_id_or_name(id_or_name, **kwargs)
    except ValueError as error:
        abort(404, str(error))


@roles_blueprint.route('/<id_or_name>', methods=['DELETE'])
@marshal_with(EmptySchema, code=204)
@doc(
    operationId='delete_role',
    summary='Delete a role.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('delete_role')
def roles_delete(id_or_name):
    try:
        Role.delete_by_id_or_name(id_or_name)
        return ('', 204)
    except ValueError as error:
        abort(404, str(error))


###############################################################################
# Role Permission Views
###############################################################################


@roles_blueprint.route('/<id_or_name>/permissions/', methods=['GET'])
@use_kwargs(RolePermissionQuerySchema, locations=['query'])
@marshal_with(RolePermissionSchema(many=True), code=200)
@doc(
    operationId='list_role_permissions',
    summary='List role permissions.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('show_role')
def role_permissions_list(id_or_name, **kwargs):
    try:
        role = Role.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))
    return Permission.query.filter_by(
        role_id=role.id).filter_in(**kwargs).all()


@roles_blueprint.route('/<id_or_name>/permissions/', methods=['POST'])
@use_kwargs(RolePermissionSchema)
@marshal_with(RolePermissionSchema(), code=201)
@doc(
    operationId='create_role_permission',
    summary='Create a role permission.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('update_role')
def role_permissions_post(id_or_name, **kwargs):
    try:
        role = Role.get_by_id_or_name(id_or_name)
        permission_name = kwargs['name']
    except ValueError as error:
        abort(404, str(error))
    try:
        return role.add_permission(permission_name), 201
    except (DuplicateError, ValueError) as error:
        abort(400, str(error))


@roles_blueprint.route(
    '/<id_or_name>/permissions/<permission_id_or_name>', methods=['GET'])
@marshal_with(RolePermissionSchema(), code=200)
@doc(
    operationId='show_role_permission',
    summary='Show a role permission.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('show_role')
def role_permissions_get(id_or_name, permission_id_or_name):
    try:
        role = Role.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))
    try:
        return Permission.get_by_role_and_id_or_name(
            role, permission_id_or_name)
    except ValueError as error:
        abort(404, str(error))


@roles_blueprint.route(
    '/<id_or_name>/permissions/<permission_id_or_name>', methods=['PATCH'])
@use_kwargs(RolePermissionUpdateSchema)
@marshal_with(RolePermissionSchema(), code=200)
@doc(
    operationId='update_role_permission',
    summary='Update a role permission.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('update_role')
def role_permissions_patch(id_or_name, permission_id_or_name, **kwargs):
    try:
        role = Role.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))
    try:
        return Permission.update_by_role_and_id_or_name(
            role, permission_id_or_name, **kwargs)
    except ValueError as error:
        abort(404, str(error))


@roles_blueprint.route(
    '/<id_or_name>/permissions/<permission_id_or_name>', methods=['DELETE'])
@marshal_with(EmptySchema, code=204)
@doc(
    operationId='delete_role_permission',
    summary='Delete a role permission.',
    description='',
    tags=['roles'],
    security=[{'api_key': []}])
@need_permission('update_role')
def role_permissions_delete(id_or_name, permission_id_or_name):
    try:
        role = Role.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))
    try:
        Permission.delete_by_role_and_id_or_name(role, permission_id_or_name)
        return ('', 204)
    except ValueError as error:
        abort(404, str(error))



from opsy.auth.utils import (login, logout, create_token, verify_token,
                             load_user, load_user_from_request)
from itsdangerous import TimedJSONWebSignatureSerializer as Serializer
from flask import current_app
from datetime import datetime


def test_login(test_user, disabled_user):
    # Test that test_user can login
    assert login('test', 'weakpass') == test_user
    # Test that invalid username is rejected
    assert login('invalid_user', 'password') is False
    # Test that invalid password is rejected
    assert login('test', 'password') is False
    # Test invalid user cannot login
    assert login('disabled', 'banhammer') is False


def test_local_login(test_user, disabled_user):
    # Test that test_user can login
    assert login('test', 'weakpass') == test_user
    # Now let's set this to be an ldap user and make sure it rejects it
    test_user.update(ldap_user=True)
    assert login('test', 'weakpass') is False
    test_user.update(ldap_user=False)
    # Test that invalid username is rejected
    assert login('invalid_user', 'password') is False
    # Test that invalid password is rejected
    assert login('test', 'password') is False
    # Test invalid user cannot login
    assert login('disabled', 'banhammer') is False


def test_logout(test_user):
    # Test that logout is processed successfully
    create_token(test_user)
    logout(test_user)
    assert test_user.session_token is None
    # Test that logout handles empty input
    assert logout('') is None


def test_create_token(test_user):
    assert test_user.session_token is None
    assert test_user.session_token_expires_at is None
    # Test that a token is created successfully
    create_token(test_user)
    assert test_user.session_token is not None
    assert test_user.session_token_expires_at is not None
    # Testing force_renew returns as expected
    assert create_token(test_user, force_renew=True) == test_user


def test_verify_token(test_user):
    # Test data
    expired_serializer = Serializer(
        current_app.config['SECRET_KEY'], expires_in=-1)
    expired_token = expired_serializer.dumps(
        {'id': test_user.id}).decode('ascii')
    bad_serializer = Serializer('12345')
    bad_token = bad_serializer.dumps({'id': test_user.id}).decode('ascii')
    good_serializer = Serializer(
        current_app.config['SECRET_KEY'],
        expires_in=current_app.config.opsy['auth']['session_token_ttl'])
    bad_id_token = good_serializer.dumps({'id': '12345'}).decode('ascii')
    # Testing the null case
    assert verify_token(None) is None
    # Test verifying with no token
    assert verify_token(test_user) is None
    # Test verifying with expired token
    test_user.update(session_token=expired_token)
    assert verify_token(test_user) is None
    # Test verifying with bad token
    test_user.update(session_token=bad_token)
    assert verify_token(test_user) is None
    # Test reducing the TTL
    create_token(test_user)
    current_ttl = current_app.config.opsy['auth']['session_token_ttl']
    current_app.config.opsy['auth']['session_token_ttl'] = -1
    assert verify_token(test_user) is None
    current_app.config.opsy['auth']['session_token_ttl'] = current_ttl
    # Test the happy path before we mess with the session token
    assert verify_token(test_user) == test_user
    # Test mismatch in token id and user.id
    orig_token = test_user.session_token
    test_user.update(session_token=bad_id_token)
    assert verify_token(test_user) is None
    test_user.update(session_token=orig_token)
    # Test mismatch in token expiration times
    test_user.update(session_token_expires_at=datetime.now())
    assert verify_token(test_user) is None


def test_load_user(test_user):
    create_token(test_user)
    assert load_user(test_user.session_token) == test_user


def test_load_user_from_request(test_user):
    class testRequest():
        def __init__(self, headers={}):
            self.headers = headers

    create_token(test_user)
    req = testRequest(headers={'X-AUTH-TOKEN': test_user.session_token})
    assert load_user_from_request(req) == test_user



import logging
import os
import pytest
import requests
from flask import Flask
from marshmallow import ValidationError
from opsy.config import load_config
from opsy.exceptions import NoConfigFile

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))


def test_app_created(app):
    """Make sure create_app isn't short changing us."""
    assert isinstance(app, Flask)


def test_swagger_json(client):
    """Make sure we have our swagger spec available."""
    response = client.get('/docs/swagger.json')
    assert response.status_code == 200
    assert b"It's Opsy!" in response.data


def test_swagger_ui(client):
    """Make sure the swagger UI is also there."""
    response = client.get('/docs/')
    assert response.status_code == 200
    assert b"Swagger UI" in response.data


def test_config(app):
    """Make sure we load in config and whatnot."""
    # Make sure it yells at us if the config file isn't found
    with pytest.raises(NoConfigFile):
        load_config(
            f'{CURRENT_DIR}/../data/not_real_opsy_config.toml')
    # Make sure it yells if the DB connection string isn't provided.
    with pytest.raises(ValidationError, match=r'.*database_uri.*'):
        load_config(
            f'{CURRENT_DIR}/../data/fake_opsy_config_no_db.toml')
    # Make sure it yells if no secret key is provided.
    with pytest.raises(ValidationError, match=r'.*secret_key.*'):
        load_config(
            f'{CURRENT_DIR}/../data/fake_opsy_config_no_secret_key.toml')
    # Make sure it yells if we don't have an app section.
    with pytest.raises(ValidationError, match=r'.*app.*'):
        load_config(
            f'{CURRENT_DIR}/../data/fake_opsy_config_no_section.toml')
    # Make sure it yells if we don't have and ldap_host with ldap_enabled.
    with pytest.raises(ValidationError, match=r'.*ldap_host.*'):
        load_config(
            f'{CURRENT_DIR}/../data/fake_opsy_config_no_ldap_host.toml')
    # Make sure it yells at us if we enable ssl but don't give cert or key.
    with pytest.raises(ValidationError, match=r'.*ssl_enabled.*'):
        load_config(
            f'{CURRENT_DIR}/../data/fake_opsy_config_no_ssl.toml')
    # Make sure our flask config makes it to the app.
    config = load_config(f'{CURRENT_DIR}/../test_opsy_config.toml')
    assert app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] is False
    assert app.config['SECRET_KEY'] == config['app']['secret_key']
    assert app.config['SQLALCHEMY_DATABASE_URI'] == \
        config['app']['database_uri']


def test_logging(app, caplog):
    """Make sure logging is working."""
    app_logger = logging.getLogger('opsy')
    app_logger.info('this is a test message')
    assert 'test message' in caplog.text
    with open('/tmp/opsy.log', 'r') as opsy_log_file:
        assert 'test message' in opsy_log_file.readlines()[-1]
    access_logger = logging.getLogger('opsy_access')
    access_logger.info('this is a test access message')
    assert 'test access message' in caplog.text
    with open('/tmp/opsy_access.log', 'r') as opsy_access_log_file:
        assert 'test access message' in opsy_access_log_file.readlines()[-1]


def test_server(test_server):
    """Test the wsgi server."""
    config = load_config(f'{CURRENT_DIR}/../test_opsy_config.toml')
    host = config['server']['host']
    port = config['server']['port']
    response = requests.get(f'http://{host}:{port}/docs/swagger.json')
    assert response.status_code == 200
    assert "It's Opsy!" in response.text



import pytest
from opsy.auth.models import User, Permission, RoleMappings


def test_user_model(test_user, test_users):
    """Test User to make sure it works correctly."""
    # Test creating a token and lookup by token
    token = test_user.get_id()
    assert token == test_user.session_token
    assert User.get_by_token(token) == test_user
    assert User.get_by_token('thisisgarbagedata') is None
    # Test is_active
    test_user.update(enabled=False)
    assert test_user.is_active == test_user.enabled == 0
    test_user.update(enabled=True)
    assert test_user.is_active == test_user.enabled == 1
    # Test that verify password returns false when we don't have a password
    test_user.update(password_hash=None)
    assert test_user.verify_password(None) is False
    assert test_user.verify_password('testing123') is False
    with pytest.raises(AttributeError, match=r'.*not a readable.*'):
        test_user.password
    test_user.update(password='testing123')
    assert test_user.verify_password('testing123') is True
    assert test_user.verify_password('testing123567') is False


def test_role_model(test_user, test_users, test_role, test_roles):
    """Test Role to make sure it works correctly."""
    # Test add_user
    test_role.add_user(test_user)
    assert test_user in test_role.users
    assert RoleMappings.query.filter_by(
        user_id=test_user.id, role_id=test_role.id).first() is not None
    # Make sure it yells at us if the user is already in the role
    with pytest.raises(ValueError):
        test_role.add_user(test_user)
    # Test remove_user
    test_role.remove_user(test_user)
    assert test_user not in test_role.users
    # Make sure it yells at us if the user isn't in the role
    with pytest.raises(ValueError):
        test_role.remove_user(test_user)
    # Test add_permission
    test_permission = test_role.add_permission('permission-to-kill-9')
    assert test_permission in test_role.permissions
    assert Permission.query.filter_by(
        role_id=test_role.id, name='permission-to-kill-9').first() == \
        test_permission
    # Make sure it yells at us if the permission already exists
    with pytest.raises(ValueError):
        test_role.add_permission('permission-to-kill-9')
    # Let's check the permission __repr__
    assert test_permission.__repr__() == \
        f'<{test_permission.__class__.__name__} {test_permission.name}>'
    # Let's add the user back and make sure they see the permission too
    test_role.add_user(test_user)
    assert test_permission in test_user.permissions
    # Test remove_permission
    test_role.remove_permission('permission-to-kill-9')
    assert 'permission-to-kill-9' not in \
        [x.name for x in test_role.permissions]
    assert Permission.query.filter_by(
        role_id=test_role.id, name='permission-to-kill-9').first() is None
    # Make sure it yells at us if the permission doesn't exists
    with pytest.raises(ValueError):
        test_role.remove_permission('permission-to-kill-9')
    # And make sure the user also doesn't see it
    assert 'permission-to-kill-9' not in \
        [x.name for x in test_user.permissions]
    # Let's test the secondary tables
    sad_permission = test_role.add_permission('sadbois')
    test_user_id = test_user.id
    sad_permission_id = sad_permission.id
    role_mapping_id = RoleMappings.query.filter_by(
        user_id=test_user.id, role_id=test_role.id).first().id
    assert sad_permission in test_user.permissions
    # Now we'll make sure the secondary table assets are cleaned up
    test_role.delete()
    assert User.query.filter_by(id=test_user_id).first() is not None
    assert sad_permission not in test_user.permissions
    assert RoleMappings.query.filter_by(id=role_mapping_id).first() is None
    assert Permission.query.filter_by(id=sad_permission_id).first() is None



import copy
import sys
from collections.abc import Mapping
from operator import attrgetter
from flask import current_app


def gwrap(some_string):
    """Returns green text."""
    return "\033[92m%s\033[0m" % some_string


def rwrap(some_string):
    """Returns red text."""
    return "\033[91m%s\033[0m" % some_string


def print_error(error, title=None, exit_script=True):
    if not title:
        if isinstance(error, Exception):
            title = error.__class__.__name__
        else:
            title = "Something broke"
    print("[%s] %s" % (rwrap(title), error), file=sys.stderr)
    if exit_script:
        sys.exit(1)


def print_notice(msg, title=None):
    if not title:
        title = "Notice"
    print("[%s] %s" % (gwrap(title), msg))


# pylint: disable=too-many-branches
def merge_dict(dest, upd, recursive_update=True, merge_lists=False):
    """
    Recursive version of the default dict.update.

    Merges upd recursively into dest
    If recursive_update=False, will use the classic dict.update, or fall back
    on a manual merge (helpful for non-dict types like FunctionWrapper)
    If merge_lists=True, will aggregate list object types instead of replace.
    The list in ``upd`` is added to the list in ``dest``, so the resulting list
    is ``dest[key] + upd[key]``. This behavior is only activated when
    recursive_update=True. By default merge_lists=False.
    .. versionchanged: 2016.11.6
        When merging lists, duplicate values are removed. Values already
        present in the ``dest`` list are not added from the ``upd`` list.

    Taken from https://github.com/saltstack/salt/blob/develop/salt/utils/dictupdate.py
    """
    if (not isinstance(dest, Mapping)) \
            or (not isinstance(upd, Mapping)):
        raise TypeError(
            'Cannot update using non-dict types in merge_dict()')
    updkeys = list(upd.keys())
    if not set(list(dest.keys())) & set(updkeys):
        recursive_update = False
    if recursive_update:
        for key in updkeys:
            val = upd[key]
            dest_subkey = dest.get(key, None)
            if isinstance(dest_subkey, Mapping) \
                    and isinstance(val, Mapping):
                ret = merge_dict(dest_subkey, val, merge_lists=merge_lists)
                dest[key] = ret
            elif isinstance(dest_subkey, list) and isinstance(val, list):
                if merge_lists:
                    merged = copy.deepcopy(dest_subkey)
                    merged.extend([x for x in val if x not in merged])
                    dest[key] = merged
                else:
                    dest[key] = upd[key]
            else:
                dest[key] = upd[key]
        return dest
    for k in upd:
        dest[k] = upd[k]
    return dest


def get_protected_routes(ignored_methods=None):
    if not ignored_methods:
        ignored_methods = []
    permissions = []
    rules = list(current_app.url_map.iter_rules())
    if not rules:
        return permissions

    rules = sorted(rules, key=attrgetter('endpoint'))

    rule_methods = [", ".join(sorted(rule.methods - set(ignored_methods)))
                    for rule in rules]

    for rule, method in zip(rules, rule_methods):
        view_func = current_app.view_functions[rule.endpoint]
        if getattr(view_func, '__rbac__', False):
            permissions.append(
                {
                    'endpoint': rule.rule,
                    'method': method,
                    'permission_needed': view_func.__rbac__[
                        'permission_needed']
                })
    return permissions


def get_valid_permissions():
    # We use set comprehension to de-dup.
    return sorted(list(
        {x['permission_needed'] for x in get_protected_routes()}))



from flask_login import UserMixin
from werkzeug.security import generate_password_hash, check_password_hash
from opsy.auth.utils import create_token
from opsy.flask_extensions import db
from opsy.models import AwareDateTime, BaseModel, NamedModel, TimeStampMixin

###############################################################################
# Auth models
###############################################################################


class User(UserMixin, NamedModel, TimeStampMixin, db.Model):

    __tablename__ = 'users'

    full_name = db.Column(db.String(64))
    email = db.Column(db.String(64), index=True)
    password_hash = db.Column(db.String(128))
    session_token = db.Column(db.String(255), index=True)
    session_token_expires_at = db.Column(AwareDateTime)
    enabled = db.Column(db.Boolean, default=False)
    ldap_user = db.Column(db.Boolean, default=False)
    roles = db.relationship('Role', secondary='role_mappings',
                            lazy='joined', backref='users')

    __table_args__ = (
        db.UniqueConstraint('session_token', name='sess_uc'),
    )

    def __init__(self, name, enabled=1, full_name=None, email=None,
                 ldap_user=False, password=None):
        self.name = name
        self.enabled = enabled
        self.full_name = full_name
        self.email = email
        self.ldap_user = ldap_user
        if password:
            self.password = password

    @property
    def permissions(self):
        permissions = []
        for role in self.roles:
            for permission in role.permissions:
                permissions.append(permission)
        return permissions

    def get_id(self):
        create_token(self)
        return self.session_token

    @classmethod
    def get_by_token(cls, token):
        if not token:
            return None
        user = cls.query.filter(cls.session_token == token).first()
        if not user:
            return None
        return user

    @property
    def is_active(self):
        return self.enabled

    @property
    def password(self):
        raise AttributeError('Password is not a readable attribute.')

    @password.setter
    def password(self, password):
        self.password_hash = generate_password_hash(password)
        self.session_token = None
        self.save()

    def verify_password(self, password):
        if not self.password_hash:
            return False
        return check_password_hash(self.password_hash, password)


class Role(NamedModel, TimeStampMixin, db.Model):

    __tablename__ = 'roles'

    ldap_group = db.Column(db.String(128), index=True)
    description = db.Column(db.Text)
    permissions = db.relationship('Permission', backref='role',
                                  cascade='all, delete-orphan')

    def add_user(self, user):
        if user in self.users:
            raise ValueError('User "%s" already added to role "%s".' % (
                user.name, self.name))
        self.users.append(user)
        self.save()

    def remove_user(self, user):
        if user not in self.users:
            raise ValueError('User "%s" not in role "%s".' % (
                user.name, self.name))
        self.users.remove(user)
        self.save()

    def add_permission(self, permission_name):
        if permission_name in [x.name for x in self.permissions]:
            raise ValueError('Permission "%s" already added to role "%s".' % (
                permission_name, self.name))
        permission_obj = Permission(
            role_id=self.id, name=permission_name)
        permission_obj.save()
        return permission_obj

    def remove_permission(self, permission_name):
        if permission_name not in [x.name for x in self.permissions]:
            raise ValueError('Permission "%s" not in role "%s".' % (
                permission_name, self.name))
        Permission.delete_by_role_and_id_or_name(self, permission_name)


class RoleMappings(BaseModel, TimeStampMixin, db.Model):

    __tablename__ = 'role_mappings'

    role_id = db.Column(
        db.String(36), db.ForeignKey('roles.id', ondelete='CASCADE'),
        index=True, nullable=False)
    user_id = db.Column(
        db.String(36), db.ForeignKey('users.id', ondelete='CASCADE'),
        index=True, nullable=False)

    role = db.relationship('Role')
    user = db.relationship('User')


class Permission(BaseModel, TimeStampMixin, db.Model):

    __tablename__ = 'permissions'

    role_id = db.Column(db.String(36), db.ForeignKey(
        'roles.id', ondelete='CASCADE'), index=True, nullable=False)
    name = db.Column(db.String(128), index=True, nullable=False)

    __table_args__ = (
        db.UniqueConstraint('role_id', 'name'),
    )

    @classmethod
    def get_by_role_and_id_or_name(cls, role, id_or_name):
        obj = cls.query.filter(db.and_(cls.role_id == role.id, db.or_(
            cls.name == id_or_name, cls.id == id_or_name))).first()
        if not obj:
            raise ValueError('No %s found with name or id "%s".' %
                             (cls.__name__, id_or_name))
        return obj

    @classmethod
    def delete_by_role_and_id_or_name(cls, role, id_or_name, **kwargs):
        return cls.get_by_role_and_id_or_name(
            role, id_or_name).delete(**kwargs)

    @classmethod
    def update_by_role_and_id_or_name(cls, role, id_or_name, **kwargs):
        return cls.get_by_role_and_id_or_name(
            role, id_or_name).update(**kwargs)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name}>'





class OpsyError(Exception):
    """Base class for exceptions in opsy."""


class NoConfigFile(OpsyError):
    """Config file not found."""


class NoConfigSection(OpsyError):
    """Config section not found."""


class MissingConfigOption(OpsyError):
    """Missing a required option in the config file."""


class DuplicateError(OpsyError):
    """Resource with this attribute already exists."""



import flask
import pytest
from flask_login import AnonymousUserMixin
from werkzeug.exceptions import Forbidden
from opsy.rbac import (is_logged_in, is_same_user, has_permission,
                       need_permission)


def test_is_logged_in(app, test_user):
    """Make sure the is logged in requirement works."""
    # Make sure we get false with the anon user.
    anon_user = AnonymousUserMixin()
    assert is_logged_in(anon_user) is False
    # ...but true if login is disabled
    app.config['LOGIN_DISABLED'] = True
    assert is_logged_in(anon_user) is True
    # Reset the flag
    app.config['LOGIN_DISABLED'] = False
    # Now let's use a test user and verify it's True
    assert is_logged_in(test_user) is True
    # ...but not true if we disable the account.
    test_user.update(enabled=False)
    assert is_logged_in(test_user) is False
    # ...but true if login is disabled.
    app.config['LOGIN_DISABLED'] = True
    assert is_logged_in(test_user) is True
    # Reset this since it's a session fixture.
    app.config['LOGIN_DISABLED'] = False


def test_is_same_user(app, test_user, mocker):
    """Make sure the is_same_user requirement works."""
    # First we need to mock the request object.
    view_args_mock = mocker.patch.object(flask.request, 'view_args')
    # Now this should return true
    view_args_mock.get.return_value = 'test'
    assert is_same_user(test_user) is True
    # ...but not this
    view_args_mock.get.return_value = 'not_test'
    assert is_same_user(test_user) is False


def test_has_permission(app, test_user, test_role):
    """Make sure has_permission works as expected."""
    # First we setup the test permission and an anon user
    test_permission = has_permission('test')
    anon_user = AnonymousUserMixin()
    # Let's disable login, these should both pass
    app.config['LOGIN_DISABLED'] = True
    assert test_permission(anon_user) is True
    assert test_permission(test_user) is True
    # Reset that and test again
    app.config['LOGIN_DISABLED'] = False
    # This should fail for the anon_user since base_permissions is empty
    assert test_permission(anon_user) is False
    # The test user isn't in a role so this should fail
    assert test_permission(test_user) is False
    # Now let's add add test to the base_permissions
    app.config.opsy['auth']['base_permissions'] = ['test']
    # This should now pass for both users
    assert test_permission(anon_user) is True
    assert test_permission(test_user) is True
    # Go ahead and reset that and test again with logged_in_permissions
    app.config.opsy['auth']['base_permissions'] = []
    app.config.opsy['auth']['logged_in_permissions'] = ['test']
    # This should now only pass for the test user
    assert test_permission(anon_user) is False
    assert test_permission(test_user) is True
    # Now let's reset that and add the user to a role with this permission
    app.config.opsy['auth']['logged_in_permissions'] = []
    test_role.add_user(test_user)
    test_role.add_permission('test')
    # This should pass since the user is in a role with this permission
    assert test_permission(test_user) is True


def test_need_permission(app, test_user, test_role):
    """Make sure the need_permission decorator works as expected."""

    # Let's make a test function and decorate it
    def test_func():
        return 'test function'

    test_decorated = need_permission('test')(test_func)
    # The resulting wrapped function should have an __rbac__ attribute with
    # the permission name in it.
    assert test_decorated.__rbac__['permission_needed'] == 'test'
    # Make sure it raises a forbidden since it's not aware of a user.
    with pytest.raises(Forbidden):
        test_decorated()
    # Now make it with on_fail. We don't really use it, but it should be
    # tested. This will just use the callable you provide it instead of raising
    # an exception.
    test_decorated = need_permission('test', on_fail=lambda: 'blah')(test_func)
    assert test_decorated() == 'blah'
    # Now let's recreate the decorated function with a user passed to it.
    # Usually you wouldn't pass the identity in like this, but it saves us
    # from having to mock things out for testing.
    test_decorated = need_permission('test', identity=test_user)(test_func)
    # Let's first make sure it doesn't allow this since our user doesn't have
    # the permission.
    with pytest.raises(Forbidden):
        test_decorated()
    # Now let's give to user permission and make sure it allows it
    test_role.add_user(test_user)
    test_role.add_permission('test')
    assert test_decorated() == 'test function'
    # Now let's give it another requirement with a permission the user doesn't
    # have
    test_decorated = need_permission(
        'test2', is_logged_in, identity=test_user)(test_func)
    # This should pass since the user is "logged in"
    assert test_decorated() == 'test function'
    # That last one will throw deprecation warnings until this is merged:
    # https://github.com/justanr/flask-allows/pull/45



import copy
import pytest
from flask import Flask
from opsy.utils import (gwrap, rwrap, print_error, print_notice, merge_dict,
                        get_protected_routes, get_valid_permissions)
from opsy.rbac import need_permission


def test_print_utils(capsys):
    """Testing to make sure our print utils work."""
    # Make sure we can print green and red text
    assert gwrap('test') == '\033[92mtest\033[0m'
    assert rwrap('test') == '\033[91mtest\033[0m'
    # Tests for print_notice
    print_notice('test message')
    captured = capsys.readouterr()
    assert captured.out == '[\033[92mNotice\033[0m] test message\n'
    print_notice('test message', title='test')
    captured = capsys.readouterr()
    assert captured.out == '[\033[92mtest\033[0m] test message\n'
    # Tests for print_error
    with pytest.raises(SystemExit) as pytest_wrapped_e:
            print_error('error')
    captured = capsys.readouterr()
    assert captured.err == '[\033[91mSomething broke\033[0m] error\n'
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1
    with pytest.raises(SystemExit) as pytest_wrapped_e:
            print_error('error', title='uhoh')
    captured = capsys.readouterr()
    assert captured.err == '[\033[91muhoh\033[0m] error\n'
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1
    print_error('error', title='uhoh', exit_script=False)
    captured = capsys.readouterr()
    assert captured.err == '[\033[91muhoh\033[0m] error\n'
    print_error(AssertionError('test'), exit_script=False)
    captured = capsys.readouterr()
    assert captured.err == '[\033[91mAssertionError\033[0m] test\n'


def test_merge_dict():
    """Make sure merge dict is working right.

    Just stealing the tests for dictupdate from salt since that's where this
    function originally came from:
    https://github.com/saltstack/salt/blob/2019.2/tests/unit/utils/test_dictupdate.py
    """
    dict1 = {'A': 'B', 'C': {'D': 'E', 'F': {'G': 'H', 'I': 'J'}}}

    # level 1 value changes
    mdict = copy.deepcopy(dict1)
    mdict['A'] = 'Z'
    res = merge_dict(copy.deepcopy(dict1), {'A': 'Z'})
    assert res == mdict

    # level 1 value changes (list replacement)
    mdict = copy.deepcopy(dict1)
    mdict['A'] = [1, 2]
    res = merge_dict(copy.deepcopy(mdict), {'A': [2, 3]},
                     merge_lists=False)
    mdict['A'] = [2, 3]
    assert res == mdict

    # level 1 value changes (list merge)
    mdict = copy.deepcopy(dict1)
    mdict['A'] = [1, 2]
    res = merge_dict(copy.deepcopy(mdict), {'A': [3, 4]},
                     merge_lists=True)
    mdict['A'] = [1, 2, 3, 4]
    assert res == mdict

    # level 1 value changes (list merge, remove duplicates, preserve order)
    mdict = copy.deepcopy(dict1)
    mdict['A'] = [1, 2]
    res = merge_dict(copy.deepcopy(mdict), {'A': [4, 3, 2, 1]},
                     merge_lists=True)
    mdict['A'] = [1, 2, 4, 3]
    assert res == mdict

    # level 2 value changes
    mdict = copy.deepcopy(dict1)
    mdict['C']['D'] = 'Z'
    res = merge_dict(copy.deepcopy(dict1), {'C': {'D': 'Z'}})
    assert res == mdict

    # level 2 value changes (list replacement)
    mdict = copy.deepcopy(dict1)
    mdict['C']['D'] = ['a', 'b']
    res = merge_dict(copy.deepcopy(mdict), {'C': {'D': ['c', 'd']}},
                     merge_lists=False)
    mdict['C']['D'] = ['c', 'd']
    assert res == mdict

    # level 2 value changes (list merge)
    mdict = copy.deepcopy(dict1)
    mdict['C']['D'] = ['a', 'b']
    res = merge_dict(copy.deepcopy(mdict), {'C': {'D': ['c', 'd']}},
                     merge_lists=True)
    mdict['C']['D'] = ['a', 'b', 'c', 'd']
    assert res == mdict

    # level 2 value changes (list merge, remove duplicates, preserve order)
    mdict = copy.deepcopy(dict1)
    mdict['C']['D'] = ['a', 'b']
    res = merge_dict(copy.deepcopy(mdict), {'C': {'D': ['d', 'c', 'b', 'a']}},
                     merge_lists=True)
    mdict['C']['D'] = ['a', 'b', 'd', 'c']
    assert res == mdict

    # level 3 value changes
    mdict = copy.deepcopy(dict1)
    mdict['C']['F']['G'] = 'Z'
    res = merge_dict(
        copy.deepcopy(dict1),
        {'C': {'F': {'G': 'Z'}}}
    )
    assert res == mdict

    # level 3 value changes (list replacement)
    mdict = copy.deepcopy(dict1)
    mdict['C']['F']['G'] = ['a', 'b']
    res = merge_dict(copy.deepcopy(mdict), {'C': {'F': {'G': ['c', 'd']}}},
                     merge_lists=False)
    mdict['C']['F']['G'] = ['c', 'd']
    assert res == mdict

    # level 3 value changes (list merge)
    mdict = copy.deepcopy(dict1)
    mdict['C']['F']['G'] = ['a', 'b']
    res = merge_dict(copy.deepcopy(mdict), {'C': {'F': {'G': ['c', 'd']}}},
                     merge_lists=True)
    mdict['C']['F']['G'] = ['a', 'b', 'c', 'd']
    assert res == mdict

    # level 3 value changes (list merge, remove duplicates, preserve order)
    mdict = copy.deepcopy(dict1)
    mdict['C']['F']['G'] = ['a', 'b']
    res = merge_dict(copy.deepcopy(mdict),
                     {'C': {'F': {'G': ['d', 'c', 'b', 'a']}}},
                     merge_lists=True)
    mdict['C']['F']['G'] = ['a', 'b', 'd', 'c']
    assert res == mdict

    # replace a sub-dictionary
    mdict = copy.deepcopy(dict1)
    mdict['C'] = 'Z'
    res = merge_dict(copy.deepcopy(dict1), {'C': 'Z'})
    assert res == mdict

    # add a new scalar value
    mdict = copy.deepcopy(dict1)
    mdict['Z'] = 'Y'
    res = merge_dict(copy.deepcopy(dict1), {'Z': 'Y'})
    assert res == mdict

    # add a dictionary
    mdict = copy.deepcopy(dict1)
    mdict['Z'] = {'Y': 'X'}
    res = merge_dict(copy.deepcopy(dict1), {'Z': {'Y': 'X'}})
    assert res == mdict

    # add a nested dictionary
    mdict = copy.deepcopy(dict1)
    mdict['Z'] = {'Y': {'X': 'W'}}
    res = merge_dict(
        copy.deepcopy(dict1),
        {'Z': {'Y': {'X': 'W'}}}
    )
    assert res == mdict

    # Now we'll do some additional testing to get this to 100% coverage
    with pytest.raises(TypeError):
        merge_dict([], [])

    mdict = {'test': None}
    res = merge_dict({}, {'test': None})
    assert res == mdict


def test_get_protected_routes():
    """Make sure get protected routes is working right."""
    # Make an empty app with no routes
    app = Flask('test_app', static_folder=None)
    # And make sure we get an empty response
    with app.app_context():
        assert get_protected_routes() == []

    # Let's add an unprotected route and make sure it's the same output

    @app.route('/test_route')
    def test_route():
        pass

    with app.app_context():
        assert get_protected_routes() == []

    # Now let's add a protected route and make sure that's in there

    @app.route('/test_protected_route')
    @need_permission('test_protected_show')
    def test_protected_route():
        pass

    expected_output = [{
        'endpoint': '/test_protected_route',
        'method': 'GET, HEAD, OPTIONS',
        'permission_needed': 'test_protected_show'
    }]
    with app.app_context():
        assert expected_output == get_protected_routes()

    # Now with a method filter
    expected_output = [{
        'endpoint': '/test_protected_route',
        'method': 'GET',
        'permission_needed': 'test_protected_show'
    }]
    with app.app_context():
        assert expected_output == get_protected_routes(
            ignored_methods=["HEAD", "OPTIONS"])


def test_get_valid_permissions():
    """Make sure get valid permissions is working right."""
    # Make an empty app with no routes
    app = Flask('test_app', static_folder=None)
    # And make sure we get an empty response
    with app.app_context():
        assert get_valid_permissions() == []

    # Let's add an unprotected route and make sure it's the same output

    @app.route('/test_route')
    def test_route():
        pass

    with app.app_context():
        assert get_valid_permissions() == []

    # Now let's add a protected route and make sure that's in there

    @app.route('/test_protected_route')
    @need_permission('test_protected_show')
    def test_protected_route():
        pass

    expected_output = ['test_protected_show']
    with app.app_context():
        assert expected_output == get_valid_permissions()



from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import validates
from opsy.exceptions import DuplicateError
from opsy.flask_extensions import db
from opsy.models import BaseModel, NamedModel, OpsyQuery, TimeStampMixin
from opsy.utils import merge_dict

###############################################################################
# Inventory models
###############################################################################


class Zone(NamedModel, TimeStampMixin, db.Model):

    __tablename__ = 'zones'

    vars = db.Column(MutableDict.as_mutable(db.JSON))
    description = db.Column(db.Text)
    hosts = db.relationship('Host', cascade='all, delete-orphan',
                            backref='zone', lazy='joined',
                            query_class=OpsyQuery)
    groups = db.relationship('Group', cascade='all, delete-orphan',
                             backref='zone', lazy='joined',
                             query_class=OpsyQuery)


class Host(NamedModel, TimeStampMixin, db.Model):

    __tablename__ = 'hosts'

    zone_id = db.Column(
        db.String(36), db.ForeignKey('zones.id', ondelete='CASCADE'),
        index=True, nullable=False)
    vars = db.Column(MutableDict.as_mutable(db.JSON))

    groups = db.relationship('Group',
                             order_by='host_group_mappings.c.priority',
                             secondary='host_group_mappings',
                             lazy='joined', backref='hosts')

    def add_group(self, group, priority=None):
        return group.add_host(self, priority=priority)

    def remove_group(self, group):
        group.remove_host(self)

    def get_group_priority(self, group):
        return group.get_host_priority(self)

    def change_group_priority(self, group, priority):
        group.change_host_priority(self, priority)

    @property
    def compiled_vars(self):
        return self.compile_vars()

    def compile_vars(self):
        compiled_dict = {}
        if self.zone.vars:
            merge_dict(compiled_dict, self.zone.vars, merge_lists=True)
        for group in self.groups:
            if group.compiled_vars:
                merge_dict(
                    compiled_dict, group.compile_vars(include_zone=False),
                    merge_lists=True)
        if self.vars:
            merge_dict(compiled_dict, self.vars, merge_lists=True)
        return compiled_dict

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.zone.name}/{self.name}>'


class Group(BaseModel, TimeStampMixin, db.Model):

    __tablename__ = 'groups'

    name = db.Column(db.String(128), index=True, nullable=False)
    parent_id = db.Column(
        db.String(36), db.ForeignKey('groups.id'), index=True)
    zone_id = db.Column(
        db.String(36), db.ForeignKey('zones.id', ondelete='CASCADE'),
        index=True, nullable=True)
    default_priority = db.Column(db.BigInteger, default=100)
    vars = db.Column(MutableDict.as_mutable(db.JSON))

    children = db.relationship(
        "Group", backref=db.backref('parent', remote_side='Group.id'))

    __table_args__ = (
        db.UniqueConstraint('name', 'zone_id'),
        db.UniqueConstraint(
            'name', 'zone_id'),
        db.Index(
            'groups_name_zone_id_key_not_null', 'name', 'zone_id', unique=True,
            postgresql_where=(db.not_(zone_id.is_(None))),
            sqlite_where=(db.not_(zone_id.is_(None)))),
        db.Index(
            'groups_name_zone_id_key_null', 'name', unique=True,
            postgresql_where=(zone_id.is_(None)),
            sqlite_where=(zone_id.is_(None)))
    )

    @classmethod
    def create(cls, name, *args, **kwargs):
        try:
            return cls(*args, name=name, **kwargs).save()
        except IntegrityError:
            db.session.rollback()
            raise DuplicateError(
                f'{cls.__name__} already exists with name "{name}" in '
                'requested zone.')

    @validates('parent_id')
    def validate_parent_id(self, key, parent_id):
        parent = Group.query.filter_by(id=parent_id).first()
        if not (parent.zone == self.zone or parent.zone is None):
            raise ValueError('Parent Group must either be in the same '
                             'Zone as child Group or not in a Zone.')
        return parent_id

    @validates('zone_id')
    def validate_zone_id(self, key, zone_id):
        if self.hosts:
            raise ValueError('Cannot change the zone of a group with hosts.')
        return zone_id

    @property
    def compiled_vars(self):
        return self.compile_vars()

    def compile_vars(self, include_zone=True):
        compiled_dict = {}
        if self.zone and include_zone and self.zone.vars:
            merge_dict(compiled_dict, self.zone.vars, merge_lists=True)
        if self.parent and self.parent.vars:
            merge_dict(compiled_dict, self.parent.vars, merge_lists=True)
        if self.vars:
            merge_dict(compiled_dict, self.vars, merge_lists=True)
        return compiled_dict

    def add_host(self, host, priority=None):
        if host in self.hosts:
            raise DuplicateError(
                'Host "{host.id}" already added to group "{self.id}".')
        if host.zone != self.zone:
            raise ValueError(
                f'Host "{host.id}" not in same zone as group '
                f'"{self.id}".')
        if not priority:
            priority = self.default_priority
        return HostGroupMapping.create(
            host_id=host.id, group_id=self.id, priority=priority)

    def remove_host(self, host):
        if host not in self.hosts:
            raise ValueError(
                'Host "{host.id}" not in group "{self.id}".')
        HostGroupMapping.query.filter_by(
            host_id=host.id, group_id=self.id).first().delete()

    def get_host_priority(self, host):
        if host not in self.hosts:
            raise ValueError(
                'Host "{host.id}" not in group "{self.id}".')
        mapping = HostGroupMapping.query.filter_by(
            host_id=host.id, group_id=self.id).first()
        return mapping.priority

    def change_host_priority(self, host, priority):
        if host not in self.hosts:
            raise ValueError(
                'Host "{host.id}" not in group "{self.id}".')
        mapping = HostGroupMapping.query.filter_by(
            host_id=host.id, group_id=self.id).first()
        mapping.update(priority=priority)

    def __repr__(self):
        zone_name = self.zone.name if self.zone else "None"
        return f'<{self.__class__.__name__} {zone_name}/{self.name}>'


class HostGroupMapping(BaseModel, TimeStampMixin, db.Model):

    __tablename__ = 'host_group_mappings'

    host_id = db.Column(
        db.String(36), db.ForeignKey('hosts.id', ondelete='CASCADE'),
        index=True, nullable=False)
    group_id = db.Column(
        db.String(36), db.ForeignKey('groups.id', ondelete='CASCADE'),
        index=True, nullable=False)
    priority = db.Column(db.BigInteger, default=100)

    host = db.relationship(
        'Host',
        backref=db.backref('group_mappings', cascade='all, delete-orphan',
                           order_by='HostGroupMapping.priority'))
    group = db.relationship(
        'Group',
        backref=db.backref('host_mappings', cascade='all, delete-orphan',
                           order_by='HostGroupMapping.priority'))

    __mapper_args__ = {
        'confirm_deleted_rows': False
    }

    @property
    def host_name(self):
        return self.host.name

    @property
    def group_name(self):
        return self.group.name

    @classmethod
    def get_by_host_and_group(cls, host_id_or_name, group_id_or_name):
        obj = cls.query.join(Host).join(Group).filter(
            db.and_(
                db.or_(
                    Host.name == host_id_or_name,
                    Host.id == host_id_or_name),
                db.or_(
                    Group.name == group_id_or_name,
                    Group.id == group_id_or_name),
                Group.zone_id == Host.zone_id)
        ).first()
        if not obj:
            raise ValueError(
                f'No mapping found with host name or id of '
                f'"{host_id_or_name}" and group name or id of '
                f'"{group_id_or_name}".')
        return obj



import uuid
from datetime import datetime, timezone
from flask_sqlalchemy import BaseQuery
from sqlalchemy import or_
from sqlalchemy.orm.attributes import CollectionAttributeImpl
from sqlalchemy.orm.base import _entity_descriptor
from opsy.flask_extensions import db
from opsy.exceptions import DuplicateError

###############################################################################
# Base models
###############################################################################


class AwareDateTime(db.TypeDecorator):
    """Results returned as aware datetimes, not naive ones."""

    impl = db.DateTime

    def process_result_value(self, value, dialect):
        if not isinstance(value, datetime):
            return value
        return value.replace(tzinfo=timezone.utc)


class TimeStampMixin:
    created_at = db.Column(AwareDateTime,
                           default=datetime.now(timezone.utc))
    updated_at = db.Column(AwareDateTime,
                           default=datetime.now(timezone.utc),
                           onupdate=datetime.now(timezone.utc))


class OpsyQuery(BaseQuery):

    def filter_in(self, ignore_none=False, **kwargs):
        filters = []
        joins = []
        for key, value in kwargs.items():
            local_descriptor = None  # for joins this is the local attribute
            if '___' in key:
                key, relationship_attr = key.split('___', 1)
            else:
                relationship_attr = None
            descriptor = _entity_descriptor(self._joinpoint_zero(), key)
            if relationship_attr:
                joins.append(descriptor)
                local_descriptor = descriptor
                descriptor = _entity_descriptor(descriptor, relationship_attr)
            if isinstance(value, str):
                filters.extend(self._get_filters_list(
                    descriptor, value, local_descriptor))
            else:
                filters.append(descriptor == value)
        new_self = self
        for descriptor in joins:
            new_self = new_self.outerjoin(descriptor)
        return new_self.filter(*filters)

    def _get_filters_list(self, descriptor, items, local_descriptor):

        filters_list = []
        if items:
            include, exclude, like, not_like = self._parse_filters(items)
            include_list = []
            exclude_list = []
            if include:
                include_list.append(descriptor.in_(include))
            if like:
                include_list.extend([descriptor.like(x) for x in like])
            if include_list:
                filters_list.append(or_(*include_list))
            if exclude:
                exclude_list.append(descriptor.in_(exclude))
            if not_like:
                exclude_list.extend([descriptor.like(x) for x in not_like])
            if exclude_list:
                if local_descriptor and isinstance(
                        local_descriptor.impl, CollectionAttributeImpl):
                    # If this is a join we want to also include things that
                    # don't match the join condition on negation. So like
                    # if the foreign key is null, for example.
                    filters_list.append(
                        ~local_descriptor.any(or_(*exclude_list)))
                else:
                    filters_list.append(~or_(*exclude_list))
        return filters_list

    def _parse_filters(self, items):
        item_list = items.split(',')
        # Wrap in a set to remove duplicates
        include = list({x for x in item_list
                        if not x.startswith('!') and '*' not in x})
        exclude = list({x[1:] for x in item_list
                        if x.startswith('!') and '*' not in x})
        like = list({x.replace('*', '%') for x in item_list
                     if not x.startswith('!') and '*' in x})
        not_like = list({x[1:].replace('*', '%') for x in item_list
                         if x.startswith('!') and '*' in x})
        return include, exclude, like, not_like

    def get_or_fail(self, ident):
        obj = self.get(ident)
        if obj is None:
            raise ValueError
        return obj

    def first_or_fail(self):
        obj = self.first()
        if obj is None:
            raise ValueError
        return obj


class BaseModel:

    query_class = OpsyQuery

    id = db.Column(db.String(36),  # pylint: disable=invalid-name
                   default=lambda: str(uuid.uuid4()), primary_key=True)

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def create(cls, **kwargs):
        obj = cls(**kwargs)
        return obj.save()

    @classmethod
    def get_by_id(cls, obj_id, **kwargs):
        return cls.query.get_or_fail(obj_id)

    @classmethod
    def delete_by_id(cls, obj_id, **kwargs):
        return cls.query.get_or_fail(obj_id).delete(**kwargs)

    @classmethod
    def update_by_id(cls, obj_id, **kwargs):
        return cls.query.get_or_fail(obj_id).update(**kwargs)

    def update(self, commit=True, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self.save() if commit else self

    def save(self):
        db.session.add(self)
        db.session.commit()
        return self

    def delete(self, commit=True):
        db.session.delete(self)
        if commit:
            db.session.commit()

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.id}>'


class NamedModel(BaseModel):

    name = db.Column(db.String(128), unique=True, index=True, nullable=False)

    def __init__(self, name, **kwargs):
        self.name = name
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def create(cls, name, *args, obj_class=None, **kwargs):
        if cls.query.filter_by(name=name).first():
            raise DuplicateError('%s already exists with name "%s".' % (
                cls.__name__, name))
        return cls(name, *args, **kwargs).save()

    @classmethod
    def get_by_id_or_name(cls, obj_id_or_name):
        obj = cls.query.filter(db.or_(
            cls.name == obj_id_or_name, cls.id == obj_id_or_name)).first()
        if not obj:
            raise ValueError('No %s found with name or id "%s".' %
                             (cls.__name__, obj_id_or_name))
        return obj

    @classmethod
    def delete_by_id_or_name(cls, obj_id_or_name, **kwargs):
        return cls.get_by_id_or_name(
            obj_id_or_name).delete(**kwargs)

    @classmethod
    def update_by_id_or_name(cls, obj_id_or_name, **kwargs):
        return cls.get_by_id_or_name(
            obj_id_or_name).update(**kwargs)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name}>'



from cheroot import wsgi
from cheroot.ssl.pyopenssl import pyOpenSSLAdapter


def create_server(app, host, port, threads=10, ssl_enabled=None,
                  certificate=None, private_key=None, ca_certificate=None):
    server = wsgi.Server((host, port), app, numthreads=threads)
    if ssl_enabled:
        server.ssl_adapter = pyOpenSSLAdapter(
            certificate, private_key, certificate_chain=ca_certificate)
    return server



import pytest
from opsy.inventory.models import Host, Group, Zone, HostGroupMapping
from opsy.exceptions import DuplicateError


def test_zone_model(test_inventory_bootstrap):
    """Test Zone to make sure it works correctly."""
    # Test to make sure deleting a zone nukes all assets in that zone
    west_zone = Zone.query.filter_by(name='west').first()
    west_prom_nodes_group_id = Group.query.filter_by(
        name='prom_nodes', zone_id=west_zone.id).first().id
    westprom_id = Host.query.filter_by(
        name='westprom', zone_id=west_zone.id).first().id
    westconsul_id = Host.query.filter_by(
        name='westconsul', zone_id=west_zone.id).first().id
    westprom_group_mapping_id = HostGroupMapping.query.filter_by(
        host_id=westprom_id, group_id=west_prom_nodes_group_id).first().id
    west_zone.delete()
    assert Group.query.filter_by(
        id=west_prom_nodes_group_id).first() is None
    assert Host.query.filter_by(id=westprom_id).first() is None
    assert Host.query.filter_by(id=westconsul_id).first() is None
    assert HostGroupMapping.query.filter_by(
        id=westprom_group_mapping_id).first() is None
    # Make sure the rest of the stuff is still there
    assert len(Zone.query.all()) == 2  # east and central
    assert len(Group.query.all()) == 3  # 1 global group, and 1 per zone
    assert len(Host.query.all()) == 4  # 2 hosts per zone
    assert len(HostGroupMapping.query.all()) == 2  # 1 mapping per zone


def test_host_model(test_host, test_group, test_inventory_bootstrap):
    """Test Host to make sure it works correctly."""
    # {'datacenter': 'west', 'prom_node': 'westprom'}
    west_zone = Zone.query.filter_by(name='west').first()
    east_zone = Zone.query.filter_by(name='east').first()
    prom_group = Group.query.filter_by(
        name='prom_nodes', zone_id=None).first()
    west_prom_group = Group.query.filter_by(
        name='prom_nodes', zone_id=west_zone.id).first()
    east_prom_group = Group.query.filter_by(
        name='prom_nodes', zone_id=east_zone.id).first()
    west_default_group = Group.query.filter_by(
        name='default', zone_id=west_zone.id).first()
    west_default_group = Group.query.filter_by(
        name='default', zone_id=west_zone.id).first()
    test_host.update(zone_id=west_zone.id)
    # Test __repr__
    assert test_host.__repr__() == (
        f'<{test_host.__class__.__name__} {test_host.zone.name}/'
        f'{test_host.name}>')
    # Test that compiled_vars should only be the zone vars at the moment
    assert test_host.compiled_vars == west_zone.vars
    # Now add vars and make sure they merge
    test_host.update(vars={'test': True})
    assert test_host.compiled_vars == {'datacenter': 'west',
                                       'prom_node': 'westprom',
                                       'test': True}
    # Test adding a group
    # Make sure it yells at us if we're not in a zone
    with pytest.raises(ValueError):
        test_host.add_group(test_group)
    # Make sure it yells at us if we're in a different zone
    test_group.update(zone_id=east_zone.id)
    with pytest.raises(ValueError):
        test_host.add_group(test_group)
    # Okay, let's actually add it now and make sure it looks right
    test_group.update(zone_id=test_host.zone.id)
    test_host.add_group(test_group)
    test_group_mapping = HostGroupMapping.query.filter_by(
        host_id=test_host.id, group_id=test_group.id).first()
    assert test_group_mapping is not None
    assert isinstance(test_group_mapping, HostGroupMapping)
    assert test_group_mapping in test_host.group_mappings
    assert test_group_mapping in test_group.host_mappings
    assert test_group_mapping.host_name == test_host.name
    assert test_group_mapping.group_name == test_group.name
    assert test_group_mapping.priority == 100
    assert test_host.get_group_priority(test_group) == 100
    # Let's also test HostGroupMapping.get_by_host_and_group works
    assert HostGroupMapping.get_by_host_and_group(
        test_host.id, test_group.id) == test_group_mapping
    assert HostGroupMapping.get_by_host_and_group(
        test_host.name, test_group.name) == test_group_mapping
    # And make sure it yells at us if we give it nonsense
    with pytest.raises(ValueError):
        HostGroupMapping.get_by_host_and_group('notarealhost', 'notarealgroup')
    # Make sure we can't change the group's zone after we've added a host
    with pytest.raises(ValueError):
        test_group.update(zone_id=None)
    # Make sure it yells at us if we try adding it again
    with pytest.raises(DuplicateError):
        test_host.add_group(test_group)
    # Test that adding vars to the group properly overides zone vars
    test_group.update(vars={'prom_node': 'truewestprom', 'shutup': 'sam'})
    assert test_host.compiled_vars == {'datacenter': 'west',
                                       'prom_node': 'truewestprom',
                                       'test': True,
                                       'shutup': 'sam'}
    # Let's add another group with a higher priority
    test_host.add_group(west_default_group, priority=110)
    west_default_group_mapping = HostGroupMapping.query.filter_by(
        host_id=test_host.id, group_id=west_default_group.id).first()
    assert west_default_group_mapping.priority == 110
    assert test_host.get_group_priority(west_default_group) == 110
    # Make sure it orders correctly (lower priority == lower index)
    assert test_host.groups == [test_group, west_default_group]
    assert test_host.group_mappings == [test_group_mapping,
                                        west_default_group_mapping]
    # Let's make sure those vars also show up
    assert test_host.compiled_vars == {'datacenter': 'west',
                                       'prom_node': 'truewestprom',
                                       'test': True,
                                       'shutup': 'meg',
                                       'default': True}
    # Let's change the priority and check again
    test_host.change_group_priority(west_default_group, 90)
    assert west_default_group_mapping.priority == 90
    assert test_host.get_group_priority(west_default_group) == 90
    assert test_host.groups == [west_default_group, test_group]
    assert test_host.group_mappings == [west_default_group_mapping,
                                        test_group_mapping]
    assert test_host.compiled_vars == {'datacenter': 'west',
                                       'prom_node': 'truewestprom',
                                       'test': True,
                                       'shutup': 'sam',
                                       'default': True}
    # Let's add a parent group in the same zone and check that the variables
    # get applied
    west_default_group.update(parent_id=west_prom_group.id)
    assert test_host.compiled_vars == {'datacenter': 'west',
                                       'prom_node': 'truewestprom',
                                       'prom_region': 'west',
                                       'test': True,
                                       'shutup': 'sam',
                                       'default': True}
    # Now a parent with no zone
    west_default_group.update(parent_id=prom_group.id)
    assert test_host.compiled_vars == {'datacenter': 'west',
                                       'prom_node': 'truewestprom',
                                       'thanos': True,
                                       'test': True,
                                       'shutup': 'sam',
                                       'default': True}
    # Now let's make sure it yells at us if we try to change the parent to a
    # group in a different zone
    with pytest.raises(ValueError):
        west_default_group.update(parent_id=east_prom_group.id)
    # Now let's remove the group
    test_host.remove_group(west_default_group)
    assert HostGroupMapping.query.filter_by(
        host_id=test_host.id, group_id=west_default_group.id).first() is None
    assert west_default_group not in test_host.groups
    assert test_host.compiled_vars == {'datacenter': 'west',
                                       'prom_node': 'truewestprom',
                                       'test': True,
                                       'shutup': 'sam'}
    # And make sure it yells at us if we try to remove it again
    with pytest.raises(ValueError):
        test_host.remove_group(west_default_group)
    # Or check the priority
    with pytest.raises(ValueError):
        test_host.get_group_priority(west_default_group)
    # Or change the priority
    with pytest.raises(ValueError):
        test_host.change_group_priority(west_default_group, 100)


def test_group_model(test_host, test_group, test_inventory_bootstrap):
    """Test Group to make sure it works correctly.

    We already tested just about all the methods through the Host tests
    since the Host methods just call the Group methods for manipulating
    group mappings."""

    # Test __repr__ without zone
    assert test_group.__repr__() == (
        f'<{test_group.__class__.__name__} None/{test_group.name}>')
    test_group.update(zone_id=test_host.zone.id)
    # Test __repr__ with zone
    assert test_group.__repr__() == (
        f'<{test_group.__class__.__name__} {test_group.zone.name}/'
        f'{test_group.name}>')
    # This should work
    Group.create('my_very_own_new_group')
    # This should throw a DuplicateError
    with pytest.raises(DuplicateError):
        Group.create('my_very_own_new_group')
    # But this should work
    zone = Zone.create('my_very_own_new_zone')
    Group.create('my_very_own_new_group', zone=zone)



from opsy.exceptions import OpsyError


class OpsyMonitoringError(OpsyError):
    """Base class for exceptions in the monitoring plugin."""


class PollFailure(OpsyMonitoringError):
    """The poll failed."""


class BackendNotFound(OpsyMonitoringError):
    """Unable to load specified backend."""



from datetime import datetime
from flask import current_app
from marshmallow import EXCLUDE, pre_load
from werkzeug.urls import url_join
from opsy.flask_extensions import ma
from opsy.monitoring.backends.base import HttpPollerBackend


class SensuBackend(HttpPollerBackend):

    def __init__(self, monitoring_service, host='localhost', protocol='http',
                 port=4567, path='', interval=60, timeout=30, username=None,
                 password=None, verify_ssl=False):
        self.monitoring_service = monitoring_service
        self.protocol = protocol
        self.host = host
        self.port = port
        self.path = path
        self.interval = interval
        self.timeout = timeout
        self.username = username
        self.password = password
        self.verify_ssl = verify_ssl

    @property
    def urls(self):
        """This should return a list of generated URLs to be polled."""
        return [url_join(self.base_url, 'events')]

    def decode(self, data):
        """This performs the data decoding. Returns list of raw events."""
        events = []
        for event in SensuEventSchema(many=True).load(data[0]):
            if not event.pop('silenced', False):
                events.append(event)
        current_app.logger.debug(f'Got from sensu: {events}')
        return events


class SensuEventSchema(ma.Schema):

    class Meta:
        fields = ('host_name', 'check_name', 'status', 'command', 'output',
                  'silenced', 'occurrences', 'updated_at')
        unknown = EXCLUDE

    silenced = ma.Boolean()
    occurrences = ma.Number()
    updated_at = ma.Method(data_key='timestamp',
                           deserialize='convert_timestamp')
    host_name = ma.String()
    check_name = ma.String()
    status = ma.Method(deserialize='convert_status')
    command = ma.String(allow_none=True)
    output = ma.String(allow_none=True)

    def convert_status(self, value):
        statuses = ['ok', 'warning', 'critical']
        try:
            return statuses[value]
        except IndexError:
            return 'unknown'

    def convert_timestamp(self, value):
        try:
            return datetime.utcfromtimestamp(int(value))
        except TypeError:
            return None

    @pre_load
    def flatten(self, data):
        data['host_name'] = data['client'].get('name')
        data['check_name'] = data['check'].get('name')
        data['status'] = data['check'].get('status')
        data['command'] = data['check'].get('command')
        data['output'] = data['check'].get('output')
        return data



from datetime import datetime
from flask import current_app
from stevedore import driver
from stevedore.exception import NoMatches
from sqlalchemy.orm import reconstructor
from opsy.flask_extensions import db
from opsy.models import TimeStampMixin, OpsyQuery, NamedModel, BaseModel
from opsy.monitoring.exceptions import BackendNotFound

###############################################################################
# Monitoring models
###############################################################################


class Event(BaseModel, TimeStampMixin, db.Model):

    __tablename__ = 'monitoring_events'

    monitoring_service_id = db.Column(
        db.String(36),
        db.ForeignKey('monitoring_services.id', ondelete='CASCADE'))
    # We grab the names here just for ease of filtering with dashboards.
    monitoring_service_name = db.Column(
        db.String(128),
        db.ForeignKey('monitoring_services.name', ondelete='CASCADE',
                      onupdate='CASCADE'))
    zone_name = db.Column(
        db.String(128),
        db.ForeignKey('zones.name', ondelete='CASCADE', onupdate='CASCADE'),
        nullable=True)
    assignee_name = db.Column(
        db.String(36), db.ForeignKey('users.name'), nullable=True)
    host_id = db.Column(
        db.String(36),
        db.ForeignKey('hosts.id'), nullable=True, default=None)
    host_name = db.Column(db.String(128), index=True)
    check_name = db.Column(db.String(128), index=True)
    state = db.Column(db.String(16), index=True, default='new')
    status = db.Column(db.String(16), nullable=False)
    interval = db.Column(db.BigInteger)
    occurrences = db.Column(db.BigInteger)
    command = db.Column(db.Text)
    output = db.Column(db.Text)
    resolved = db.Column(db.Boolean(), default=False)
    resolved_at = db.Column(db.DateTime, default=None)
    extra = db.Column(db.JSON)

    monitoring_service = db.relationship('MonitoringService', backref='events',
                                         query_class=OpsyQuery,
                                         foreign_keys=[monitoring_service_id])
    zone = db.relationship('Zone', backref='events', query_class=OpsyQuery)
    host = db.relationship('Host', backref='events', query_class=OpsyQuery)
    assignee = db.relationship('User', backref='events', query_class=OpsyQuery)

    __table_args__ = (
        db.CheckConstraint(status.in_(
            ['ok', 'warning', 'critical', 'unknown'])),
        db.CheckConstraint(state.in_(
            ['new', 'acknowledged', 'resolved']))
    )

    def __init__(self, monitoring_service, **kwargs):
        self.monitoring_service_id = monitoring_service.id
        self.monitoring_service_name = monitoring_service.name
        self.zone_name = getattr(monitoring_service.zone, 'name', None)
        super().__init__(**kwargs)

    def resolve(self, commit=True):
        current_app.logger.debug(f'Resolving event: {self}')
        self.state = 'resolved'
        self.resolved = True
        self.resolved_at = datetime.now()
        return self.save() if commit else self


class MonitoringService(NamedModel, TimeStampMixin, db.Model):

    __tablename__ = 'monitoring_services'

    zone_id = db.Column(
        db.String(36), db.ForeignKey('zones.id', ondelete='CASCADE'),
        nullable=True)
    enabled = db.Column(db.Boolean(), default=False)
    backend_name = db.Column(db.String(128), default=None)
    backend_config = db.Column(db.JSON, default=None)
    interval = db.Column(db.BigInteger, default=None)
    last_poll_time = db.Column(db.DateTime, default=None)
    extra = db.Column(db.JSON)

    zone = db.relationship('Zone', backref='monitoring_services',
                           query_class=OpsyQuery)

    def __init__(self, name, zone=None, backend_name=None, backend_config=None,
                 **kwargs):
        if zone:
            self.zone_id = zone.id
        if backend_name:
            try:
                self.backend = driver.DriverManager(
                    namespace='opsy.monitoring.backend',
                    name=backend_name, invoke_on_load=True,
                    invoke_args=(self,),
                    invoke_kwds=backend_config).driver
                self.backend_config = self.backend.config
            except NoMatches:
                raise BackendNotFound(f'Unable to load backend {backend_name}')
        else:
            self.backend = None
            self.backend_config = None
        super().__init__(name, backend_name=backend_name,
                         backend_config=backend_config, **kwargs)

    @reconstructor
    def init_on_load(self):
        """
        Load instance of backend manager here.

        The idea is to have the args for creating the manager stored in the
        backend_config column in the db. That way backends have quite a bit of
        flexibility in defining their connection method.
        """
        if self.backend_name:
            try:
                self.backend = driver.DriverManager(
                    namespace='opsy.monitoring.backend',
                    name=self.backend_name, invoke_on_load=True,
                    invoke_args=(self,),
                    invoke_kwds=self.backend_config).driver
                self.backend_config = self.backend.config
            except NoMatches:
                raise BackendNotFound(
                    f'Unable to load backend {self.backend_name}')
        else:
            self.backend = None


class Dashboard(NamedModel, TimeStampMixin, db.Model):

    __tablename__ = 'monitoring_dashboards'

    description = db.Column(db.String(256))
    enabled = db.Column('enabled', db.Boolean(), default=0)
    owner = db.Column(
        db.String(36), db.ForeignKey('users.id'), nullable=True)
    zone_filter = db.Column(db.Text())
    monitoring_service_filter = db.Column(db.Text())
    host_filter = db.Column(db.Text())
    check_filter = db.Column(db.Text())

    user = db.relationship('User', backref='monitoring_services',
                           query_class=OpsyQuery)



from marshmallow import RAISE
from marshmallow import fields as ma_fields
from marshmallow_sqlalchemy import field_for
from opsy.inventory.models import Zone, Host, Group, HostGroupMapping
from opsy.flask_extensions import ma
from opsy.schema import BaseSchema, Hyperlinks

###############################################################################
# Sqlalchemy schemas
###############################################################################


class ZoneSchema(BaseSchema):

    class Meta:
        model = Zone
        fields = ('id', 'name', 'description', 'vars', 'created_at',
                  'updated_at', '_links')
        ordered = True
        unknown = RAISE

    id = field_for(Zone, 'id', dump_only=True,
                   description='The ID of the zone.')
    name = field_for(Zone, 'name', required=True,
                     description='The name of the zone.')
    vars = field_for(Zone, 'vars', field_class=ma_fields.Dict)
    created_at = field_for(Zone, 'created_at', dump_only=True)
    updated_at = field_for(Zone, 'updated_at', dump_only=True)
    _links = Hyperlinks(
        {"self": ma.URLFor("inventory_zones.zones_get", id_or_name="<id>"),
         "collection": ma.URLFor("inventory_zones.zones_list")},
        dump_only=True
    )


class ZoneUpdateSchema(ZoneSchema):

    class Meta:
        model = Zone
        fields = ('name', 'description', 'vars')
        ordered = True
        unknown = RAISE

    name = field_for(Zone, 'name', required=False,
                     description='The name of the zone.')


class ZoneQuerySchema(ZoneSchema):

    class Meta:
        model = Zone
        fields = ('id', 'name', 'description')
        ordered = True
        unknown = RAISE

    id = field_for(
        Zone, 'id',
        description='The unique UUID of the zone.')
    name = field_for(
        Zone, 'name', required=False,
        description='The unique name of the zone.')
    description = field_for(
        Zone, 'name', required=False,
        description='The description of the zone.')


class ZoneRefSchema(ZoneSchema):

    class Meta:
        model = Zone
        fields = ('id', 'name', '_links')
        ordered = True
        unknown = RAISE


class HostSchema(BaseSchema):

    class Meta:
        model = Host
        fields = ('id', 'zone_id', 'name', 'vars', 'compiled_vars',
                  'zone', 'group_mappings', 'created_at', 'updated_at',
                  '_links')
        ordered = True
        unknown = RAISE

    id = field_for(Host, 'id', dump_only=True)
    name = field_for(Host, 'name', required=True)
    zone_id = field_for(Host, 'zone_id', required=True)
    created_at = field_for(Host, 'created_at', dump_only=True)
    updated_at = field_for(Host, 'updated_at', dump_only=True)
    vars = field_for(Host, 'vars', field_class=ma_fields.Dict)
    compiled_vars = ma_fields.Dict(dump_only=True)

    zone = ma.Nested(  # pylint: disable=no-member
        'ZoneRefSchema', dump_only=True)
    group_mappings = ma.Nested(  # pylint: disable=no-member
        'HostGroupMappingRefSchema', attribute='group_mappings',
        dump_only=True, many=True)

    _links = Hyperlinks(
        {"self": ma.URLFor("inventory_hosts.hosts_get", id_or_name="<id>"),
         "collection": ma.URLFor("inventory_hosts.hosts_list")},
        dump_only=True
    )


class HostUpdateSchema(HostSchema):

    class Meta:
        model = Host
        fields = ('zone_id', 'name', 'vars')
        ordered = True
        unknown = RAISE

    name = field_for(Host, 'name', required=False)


class HostQuerySchema(HostSchema):

    class Meta:
        model = Host
        fields = ('id', 'zone_id', 'zone_name', 'group_id', 'group_name',
                  'name')
        ordered = True
        unknown = RAISE

    id = field_for(Host, 'id')
    name = field_for(Host, 'name', required=False)
    zone_id = field_for(Host, 'zone_id', required=False)
    zone_name = ma_fields.String(attribute='zone___name')
    group_id = ma_fields.String(attribute='groups___id')
    group_name = ma_fields.String(attribute='groups___name')


class HostRefSchema(HostSchema):

    class Meta:
        model = Host
        fields = ('id', 'name', '_links')
        ordered = True
        unknown = RAISE


class GroupSchema(BaseSchema):

    class Meta:
        model = Group
        fields = ('id', 'zone_id', 'parent_id', 'name', 'default_priority',
                  'vars', 'compiled_vars', 'created_at',
                  'updated_at', '_links')
        ordered = True
        unknown = RAISE

    id = field_for(Group, 'id', dump_only=True)
    name = field_for(Group, 'name', required=True)
    zone_id = field_for(Group, 'zone_id', allow_none=True)
    parent_id = field_for(Group, 'parent_id', allow_none=True)
    created_at = field_for(Group, 'created_at', dump_only=True)
    updated_at = field_for(Group, 'updated_at', dump_only=True)
    vars = field_for(Group, 'vars', field_class=ma_fields.Dict)
    compiled_vars = ma_fields.Dict(dump_only=True)
#    zone = ma.Nested(  # pylint: disable=no-member
#        'ZoneRefSchema', dump_only=True)
#    parent = ma.Nested(  # pylint: disable=no-member
#        'GroupRefSchema', dump_only=True)

    _links = Hyperlinks(
        {"self": ma.URLFor("inventory_groups.groups_get", group_id="<id>"),
         "collection": ma.URLFor("inventory_groups.groups_list")},
        dump_only=True
    )


class GroupUpdateSchema(GroupSchema):

    class Meta:
        model = Group
        fields = ('zone_id', 'parent_id', 'name', 'default_priority', 'vars')
        ordered = True
        unknown = RAISE

    name = field_for(Group, 'name', required=False)


class GroupQuerySchema(GroupSchema):

    class Meta:
        model = Group
        fields = ('id', 'name', 'zone_id', 'zone_name', 'parent_id',
                  'parent_name', 'host_id', 'host_name', 'default_priority')
        ordered = True
        unknown = RAISE

    id = field_for(Group, 'id')
    name = field_for(Group, 'name', required=False)
    zone_id = field_for(Group, 'zone_id', required=False)
    zone_name = ma_fields.String(attribute='zone___name')
    parent_name = ma_fields.String(attribute='parent___name')
    host_id = ma_fields.String(attribute='hosts___id')
    host_name = ma_fields.String(attribute='hosts___name')


class GroupRefSchema(GroupSchema):

    class Meta:
        model = Group
        fields = ('id', 'name', '_links')
        ordered = True
        unknown = RAISE


class HostGroupMappingSchema(BaseSchema):

    class Meta:
        model = HostGroupMapping
        fields = ('id', 'host_id', 'group_id', 'priority', 'host', 'group',
                  'created_at', 'updated_at', '_links')
        ordered = True
        unknown = RAISE

    id = field_for(HostGroupMapping, 'id', dump_only=True)
    host_id = field_for(HostGroupMapping, 'host_id', dump_only=True)
    created_at = field_for(HostGroupMapping, 'created_at', dump_only=True)
    updated_at = field_for(HostGroupMapping, 'updated_at', dump_only=True)
    host = ma.Nested(  # pylint: disable=no-member
        'HostRefSchema', dump_only=True)
    group = ma.Nested(  # pylint: disable=no-member
        'GroupRefSchema', dump_only=True)

    _links = Hyperlinks({
        "self": ma.URLFor("inventory_hosts.host_group_mappings_get",
                          id_or_name="<host_id>",
                          group_id_or_name="<group_id>"),
        "collection": ma.URLFor("inventory_hosts.host_group_mappings_list",
                                id_or_name="<host_id>")}, dump_only=True)


class HostGroupMappingUpdateSchema(HostGroupMappingSchema):

    class Meta:
        model = HostGroupMapping
        fields = ('priority', )
        ordered = True
        unknown = RAISE


class HostGroupMappingQuerySchema(HostGroupMappingSchema):

    class Meta:
        model = HostGroupMapping
        fields = ('id', 'group_id', 'group_name', 'priority')
        ordered = True
        unknown = RAISE

    id = field_for(HostGroupMapping, 'id')
    group_name = ma_fields.String(attribute='group___name')


class HostGroupMappingRefSchema(HostGroupMappingSchema):

    class Meta:
        model = HostGroupMapping
        fields = ('id', 'host_id', 'host_name', 'group_id', 'group_name',
                  'priority', '_links')
        ordered = True
        unknown = RAISE



from faker import Faker
from opsy.auth.models import Role, User
from opsy.utils import get_valid_permissions


def test_user():
    test_user = User.create(
        name='test', full_name='Test User', password='weakpass')
    test_role = Role.create(name='users')
    test_role.add_user(test_user)
    return test_user


def disabled_user():
    return User.create(
        name='disabled', full_name='Disabled User', password='banhammer',
        enabled=0)


def admin_user():
    admin_user = User.create(
        name='admin', full_name='Admin User', password='password123')
    admin_role = Role.create(name='admins')
    for permission in get_valid_permissions():
        admin_role.add_permission(permission)
    admin_role.add_user(admin_user)
    return admin_user


def test_users(number=10):
    fake = Faker(locale='en_US')
    fake.seed(8675309)
    users = []
    for x in range(0, number):
        users.append(User.create(name=fake.user_name(), full_name=fake.name(),
                                 password=fake.password()))
    return users


def test_role():
    test_role = Role.create(name='test')
    test_role.add_permission('test1')
    test_role.add_permission('test2')
    return test_role


def test_roles(number=10):
    fake = Faker(locale='en_US')
    fake.seed(8675309)
    roles = []
    for x in range(0, number):
        roles.append(Role.create(name=fake.domain_word()))
    return roles
